import copy
import logging
from math import sin, cos, pi, floor
import struct

from trosnoth.const import OFF_MAP_DEATH
from trosnoth.model.shot import Shot, PredictedGrenadeTrajectory
from trosnoth.model.trosball import PredictedTrosballTrajectory
from trosnoth.model.unit import Unit
from trosnoth.model.upgrades import Grenade
from trosnoth.utils.event import Event
from trosnoth.utils.math import distance, isNear

from trosnoth.messages import (
    PlayerUpdateMsg, ResyncPlayerMsg, ChatFromServerMsg, ResyncAcknowledgedMsg,
)

RESPAWN_CAMP_TIME = 1.0
MAX_STARS = 10
MAX_RESYNC_TIME = 4     # seconds

log = logging.getLogger('player')


class Player(Unit):
    '''Maintains the state of a player. This could be the user's player, a
    player on the network, or conceivably even a bot.
    '''

    HALF_WIDTH = 10
    HALF_HEIGHT = 19

    def __init__(
            self, world, nick, team, id, dead=False, bot=False,
            *args, **kwargs):
        super(Player, self).__init__(world, *args, **kwargs)

        self.world = world

        self.onDied = Event()       # (killer, deathType)
        self.onKilled = Event()     # (target, deathType)
        self.onZombieHit = Event()  # (killer, shot, deathType)
        self.onStarsChanged = Event()
        self.onNeutralisedSector = Event()  # (zoneCount)
        self.onTaggedZone = Event()         # (zone, previousOwner)
        self.onUsedUpgrade = Event()        # (upgrade)
        self.onAbandonUpgrade = Event()     # (upgrade)
        self.onUpgradeGone = Event()        # (upgrade)
        self.onShotFired = Event()          # (shot)
        self.onShotHurtPlayer = Event()     # (target, shot)
        self.onHitByShot = Event()          # (shot)
        self.onShieldDestroyed = Event()    # (shooter, deathType)
        self.onDestroyedShield = Event()    # (target, deathType)
        self.onPrivateChatReceived = Event()    # (text, sender)
        self.onGotTrosball = Event()
        self.onStarsSpent = Event()         # (stars)
        self.onRespawned = Event()
        self.onShotHitSomething = Event()   # (shot)
        self.onOverlayDebugHook = Event()   # (viewManager, screen, sprite)

        # Identity
        self.nick = nick
        self.user = None      # If authenticated, this will have a value.
        self.agent = None
        self.team = team
        self.id = id
        self.bot = bot

        # Preferences during the voting phase.
        self.preferredTeam = ''
        self.prefersTournamentTeam = False
        self.readyToStart = False
        self.readyForTournament = False
        self.preferredSize = (0, 0)
        self.preferredDuration = 0

        # Input state
        self._state = {'left':  False,
                       'right': False,
                       'jump':  False,
                       'down': False,
                       'respawn': False}

        # Physics
        self._alreadyJumped = False
        self._jumpTime = 0.0
        self.yVel = 0
        self.attachedObstacle = None
        self.motionState = 'fall'   # fall / ground / leftwall / rightwall
        self._unstickyWall = None
        self._ignore = None         # Used when dropping through a platform.
        self.angleFacing = 1.57
        self.ghostThrust = 0.0      # Determined by mouse position
        self._faceRight = True
        self.reloadTime = 0.0
        self.reloadFrom = 0.0
        self.turretHeat = 0.0
        self.respawnGauge = 0.0
        self.upgradeGauge = 0.0
        self.upgradeTotal = 0.0

        # Upgrade-related
        self.pendingUpgrade = None
        self.upgrade = None
        self.turretOverHeated = False
        self.invulnerableUntil = None
        self.mgBulletsRemaining = 0
        self.mgFiring = False

        # Other state info
        self.starsAtLastDeath = 0
        self.stars = world.PLAYER_RESET_STARS
        if dead:
            self.health = 0
        else:
            self.health = self.world.physics.playerRespawnHealth
        self.zombieHits = 0
        self.resyncing = False
        self.resyncExpiry = None

    def clone(self):
        '''
        Used to create the local representation of this player object.
        '''
        result = copy.copy(self)
        result._state = copy.copy(self._state)
        result.onDied = Event()
        result.onKilled = Event()
        result.onZombieHit = Event()
        result.onStarsChanged = Event()
        result.onNeutralisedSector = Event()
        result.onTaggedZone = Event()
        result.onUsedUpgrade = Event()
        result.onAbandonUpgrade = Event()
        result.onUpgradeGone = Event()
        result.onShotFired = Event()
        result.onShotHurtPlayer = Event()
        result.onHitByShot = Event()
        result.onShieldDestroyed = Event()
        result.onDestroyedShield = Event()
        result.onPrivateChatReceived = Event()
        result.onGotTrosball = Event()
        result.onStarsSpent = Event()
        result.onRespawned = Event()
        result.onShotHitSomething = Event()
        result.onOverlayDebugHook = Event()
        return result

    def dump(self):
        return {
            'id': self.id,
            'teamId': self.team.id if self.team else '\x00',
            'nick': self.nick,
            'bot': self.bot,
            'update': self.getPlayerUpdateArgs(),
            'upgrade':
                self.upgrade.upgradeType
                if self.upgrade else None,
            'upgradeTime': self.upgradeGauge,
            'preferences': {
                'team': self.preferredTeam,
                'tournament': self.prefersTournamentTeam,
                'size': self.preferredSize,
                'duration': self.preferredDuration,
                'ready': self.readyToStart,
                'readyForTournament': self.readyForTournament,
            }
        }

    def restore(self, data):
        team = self.world.teamWithId[data['teamId']]
        bot = data['bot']
        update = data['update']
        upgrade = data['upgrade']
        preferences = data['preferences']

        self.nick = data['nick']
        self.team = team
        self.bot = bot

        if self.upgrade and self.upgrade.upgradeType != upgrade:
            self.deleteUpgrade()
        if upgrade:
            if not self.upgrade or self.upgrade.upgradeType != upgrade:
                self.upgradeUsed(upgrade)
                self.upgradeGauge = data['upgradeTime']

        self.applyPlayerUpdate(PlayerUpdateMsg(*update))

        self.preferredTeam = preferences['team']
        self.preferredTournamentTeam = preferences['tournament']
        self.preferredSize = tuple(preferences['size'])
        self.preferredDuration = preferences['duration']
        self.readyToStart = preferences['ready']
        self.readyForTournament = preferences['readyForTournament']

    @property
    def dead(self):
        realHealth = self.health - self.zombieHits
        if self.hasProtectiveUpgrade():
            realHealth += self.upgrade.protections
        return realHealth <= 0

    @property
    def knowsItsDead(self):
        '''
        It's possible for a player to be dead and not know it (e.g. killed by
        grenade or shoxwave and hasn't yet received the network message). It
        this case the easiest way to keep the server and client in sync is by
        treating it as a ghost that moves like a living player.
        '''
        return self.health <= 0

    def isStatic(self):
        if self.oldPos != self.pos:
            return False
        if self.motionState == 'fall' and not self.knowsItsDead:
            return False
        if self.upgrade is not None:
            return False
        if self.knowsItsDead and self.respawnGauge != 0:
            return False
        if self.knowsItsDead and self.ghostThrust != 0:
            return False
        if self.reloadTime != 0:
            return False
        return True

    def getXKeyMotion(self):
        if self._state['left'] and not self._state['right']:
            return -1
        if self._state['right'] and not self._state['left']:
            return 1
        return 0

    def hasTrosball(self):
        return (
            self.world.trosballPlayer
            and self.world.trosballPlayer.id == self.id)

    @staticmethod
    def getObstacles(mapBlockDef):
        '''
        Return which obstacles in the given map block apply to this kind of
        unit.
        '''
        return mapBlockDef.obstacles + mapBlockDef.ledges

    @property
    def identifyingName(self):
        if self.user is None:
            return self.nick
        return self.user.username

    def hasProtectiveUpgrade(self):
        return self.upgrade is not None and self.upgrade.upgradeType in 'sd'

    def hasVisibleShield(self):
        if not self.hasProtectiveUpgrade():
            return False
        return self.upgrade.protections > self.zombieHits

    @property
    def phaseshift(self):
        return self.upgrade is not None and self.upgrade.upgradeType == 'h'

    @property
    def turret(self):
        return self.upgrade is not None and self.upgrade.upgradeType == 't'

    @property
    def shoxwave(self):
        return self.upgrade is not None and self.upgrade.upgradeType == 'w'

    @property
    def machineGunner(self):
        return self.upgrade is not None and self.upgrade.upgradeType in 'xd'

    @property
    def hasRicochet(self):
        return self.upgrade is not None and self.upgrade.upgradeType in 'rd'

    @property
    def ninja(self):
        return self.upgrade is not None and self.upgrade.upgradeType in 'nd'

    @property
    def disruptive(self):
        return self.upgrade is not None and self.upgrade.upgradeType == 'm'

    @property
    def bomber(self):
        return self.upgrade is not None and self.upgrade.upgradeType == 'b'

    @property
    def canMove(self):
        return not self.bomber

    @property
    def invisible(self):
        if self.ninja:
            if self.reloadTime != 0.0:
                return False
            if self.distanceFromCentre() > 200:
                return True
            zone = self.getZone()
            if zone and self.isFriendsWithTeam(zone.owner):
                return True
            return False
        return False

    @property
    def currentProjectileTrajectory(self):
        if self.dead:
            return None
        if self.hasTrosball():
            return PredictedTrosballTrajectory(self.world, self)
        if self.upgrade is not None:
            # Have an upgrade = cannot throw
            return None
        return PredictedGrenadeTrajectory(
            self.world, self, Grenade.timeRemaining)

    def distanceFromCentre(self):
        zone = self.getZone()
        if not zone:
            return 2000
        return distance(zone.defn.pos, self.pos)

    def isAttachedToWall(self):
        '''
        Returns False if this player is not attached to a wall, otherwise
        returns 'left' or 'right' to indicate whether the wall is on the
        player's left or right.
        '''
        if self.motionState == 'leftwall':
            return 'left'
        elif self.motionState == 'rightwall':
            return 'right'
        return False

    def detachFromEverything(self):
        self.setAttachedObstacle(None)

    def isOnGround(self):
        '''
        Returns True iff this player is on the ground (whether that ground is a
        ledge or solid ground).
        '''
        return self.motionState == 'ground'

    def isOnPlatform(self):
        return self.isOnGround() and self.attachedObstacle.drop

    def isFriendsWith(self, other):
        '''
        Returns True iff self and other are on the same team.
        '''
        if other.id == self.id:
            return True
        return self.isFriendsWithTeam(other.team)

    def isFriendsWithTeam(self, team):
        if team is None:
            return False
        return self.team == team

    def inRespawnableZone(self):
        zone = self.getZone()
        return self.isZoneRespawnable(zone)

    def isZoneRespawnable(self, zone):
        if not zone:
            return False
        return self.team is None or zone.owner == self.team

    @property
    def teamId(self):
        if self.team is None:
            return '\x00'
        return self.team.id

    @property
    def teamName(self):
        if self.team is None:
            return self.world.rogueTeamName
        return self.team.teamName

    def isEnemyTeam(self, team):
        '''
        Returns True iff the given team is an enemy team of this player. It is
        not enough for the team to be neutral (None), it must actually be an
        enemy for this method to return True.
        '''
        if team is None:
            return False
        return self.team != team

    def isInvulnerable(self):
        iu = self.invulnerableUntil
        if iu is None:
            return False
        elif self.world.timer.getElapsedTime() > iu:
            self.invulnerableUntil = None
            return False
        return True

    def getTeamStars(self):
        if self.team is None:
            return self.stars
        return self.world.getTeamStars(self.team)

    @property
    def isMinimapDisrupted(self):
        for team in self.world.teams:
            if team != self.team and team.usingMinimapDisruption:
                return True
        return False

    @staticmethod
    def _leftValidator(obs):
        return obs.deltaPt[0] == 0 and obs.deltaPt[1] > 0

    @staticmethod
    def _rightValidator(obs):
        return obs.deltaPt[0] == 0 and obs.deltaPt[1] < 0

    @staticmethod
    def _groundValidator(obs):
        return obs.deltaPt[0] > 0

    def setPos(self, pos, attached):
        self.pos = self.oldPos = pos

        self._ignore = None
        if attached == 'f':
            obstacle = None
        elif attached == 'g':
            obstacle, dX, dY = self.world.physics.trimPathToObstacle(
                self, 0, 10, set())
        elif attached == 'l':
            obstacle, dX, dY = self.world.physics.trimPathToObstacle(
                self, -10, 0, set())
        elif attached == 'r':
            obstacle, dX, dY = self.world.physics.trimPathToObstacle(
                self, 10, 0, set())
        self.setAttachedObstacle(obstacle)

    def __str__(self):
        return self.nick

    def getDetails(self):
        if self.id == -1:
            return {}

        pID = struct.unpack('B', self.id)[0]
        nick = self.nick
        team = self.teamId
        dead = self.dead
        stars = self.stars
        if self.upgrade:
            upgrade = str(self.upgrade)
        else:
            upgrade = None

        return {
            'pID': pID,
            'nick': nick,
            'team': team,
            'dead': dead,
            'stars': stars,
            'upgrade': upgrade,
            'bot': self.bot,
            'ready': self.readyToStart,
            'tournament': self.readyForTournament,
        }

    def removeFromGame(self):
        '''Called by network client when server says this player has left the
        game.'''

        if self.upgrade:
            self.upgrade.delete()

        if self.hasTrosball():
            self.world.trosballLost(self)

    def isSolid(self):
        return not self.knowsItsDead

    def ignoreObstacle(self, obstacle):
        return self._ignore == obstacle

    def continueOffMap(self):
        if not self.dead:
            # Player is off the map: mercy killing.
            self.killOutright(OFF_MAP_DEATH, resync=False)
        return False

    def canEnterZone(self, newZone):
        if newZone == self.getZone():
            return True
        world = self.world
        if (not world.abilities.leaveFriendlyZones
                and newZone is None and self.dead):
            # Pre-game ghost cannot enter purple zones.
            return False

        if (newZone is not None and
                not world.abilities.leaveFriendlyZones
                and newZone.owner != self.team):
            # Disallowed zone change.
            return False

        return True

    def updateState(self, key, value):
        '''Update the state of this player. State information is information
        which is needed to calculate the motion of the player. For a
        human-controlled player, this is essentially only which keys are
        pressed. Keys which define a player's state are: left, right, jump and
        down.
        Shooting is processed separately.'''

        if self.resyncing:
            return

        if key == 'left' or key == 'right':
            self._unstickyWall = None

        # Ignore messages if we already know the answer.
        if self._state[key] == value:
            return
        if key == 'jump':
            self._alreadyJumped = False

        # Set the state.
        self._state[key] = value

    def getState(self, key):
        return self._state[key]

    def processJumpState(self):
        '''
        Checks the player's jump key state to decide whether to initiate a jump
        or stop a jump.
        '''
        jumpKeyDown = self._state['jump'] and self.canMove
        if jumpKeyDown:
            if self.motionState == 'fall':
                return
            if self._alreadyJumped:
                return

            # Otherwise, initiate the jump.
            self._jumpTime = self.world.physics.playerMaxJumpTime
            if self.isAttachedToWall():
                self._unstickyWall = self.attachedObstacle
            self.detachFromEverything()
            self._alreadyJumped = True
        elif self._jumpTime > 0:
            # If we're jumping, cancel the jump.
            # The following line ensures that small jumps are possible
            #  while large jumps still curve.
            self.yVel = (
                -(1 - self._jumpTime / self.world.physics.playerMaxJumpTime)
                * self.world.physics.playerJumpThrust)
            self._jumpTime = 0

    def lookAt(self, angle, thrust=None):
        '''Changes the direction that the player is looking.  angle is in
        radians and is measured clockwise from vertical.'''

        if self.resyncing:
            return

        if thrust is not None:
            self.ghostThrust = min(1, max(0, thrust))

        angle = (angle + pi) % (2 * pi) - pi
        if self.angleFacing == angle:
            return

        self.angleFacing = angle
        self._faceRight = angle > 0

    def isFacingRight(self):
        return self._faceRight

    def upgradeUsed(self, upgradeType, local=None):
        upgradeKind = self.world.getUpgradeType(upgradeType)
        self.upgrade = upgradeKind(self)
        self.upgradeGauge = upgradeKind.timeRemaining
        self.upgradeTotal = upgradeKind.timeRemaining
        if local:
            self.upgrade.localUse(local)
        else:
            self.upgrade.use()
        self.onUsedUpgrade(self.upgrade)

    def checkUpgrades(self):
        # Update upgrade gauge based on Time:
        if self.upgrade and self.upgradeGauge > 0:
            self.upgradeGauge -= self.world.tickPeriod
            if self.upgradeGauge <= 0:
                self.upgrade.timeIsUp()
                self.deleteUpgrade()

    def updateGhost(self):
        deltaT = self.world.tickPeriod
        deltaX = (
            self.world.physics.playerMaxGhostVel * deltaT
            * sin(self.angleFacing) * self.ghostThrust)
        deltaY = (
            -self.world.physics.playerMaxGhostVel * deltaT
            * cos(self.angleFacing) * self.ghostThrust)

        self.world.physics.moveUnit(self, deltaX, deltaY)

        # Update respawn gauge based on Time:
        if self.respawnGauge >= 0:
            self.respawnGauge -= deltaT
            if self.respawnGauge < 0:
                self.respawnGauge = 0

    def updateTurret(self):
        deltaT = self.world.tickPeriod
        self.reloadTime = max(0.0, self.reloadTime - deltaT)
        self.turretHeat = max(0.0, self.turretHeat - deltaT)
        if self.turretOverHeated and self.turretHeat == 0.0:
            self.turretOverHeated = False

    def getAbsVelocity(self):
        '''
        Return the absolute value of the velocity the player should have
        (assuming the player is living), given the direction that the player is
        facing and the keys (left/right) being pressed.
        '''
        if not self.canMove:
            return 0

        # Consider horizontal movement of player.
        if self._state['left'] and not self._state['right']:
            if self._faceRight:
                return -self.world.physics.playerSlowXVel
            return -self.world.physics.playerXVel

        if self._state['right'] and not self._state['left']:
            if self._faceRight:
                return self.world.physics.playerXVel
            return self.world.physics.playerSlowXVel

        return 0

    def getCurrentVelocity(self):
        '''
        Used to decide how fast the Trosball should go if it's dropped.
        '''
        return (self.getAbsVelocity(), self.yVel)

    def dropThroughFloor(self):
        self._ignore = self.attachedObstacle
        self.detachFromEverything()

    def dropOffWall(self):
        self._unstickyWall = self.attachedObstacle
        self.detachFromEverything()

    def moveAlongGround(self, absVel):
        if absVel == 0:
            return

        deltaT = self.world.tickPeriod
        attachedObstacle = self.attachedObstacle
        for i in xrange(100):
            if deltaT <= 0:
                obstacle = None
                break

            oldPos = self.pos
            deltaX, deltaY = attachedObstacle.walkTrajectory(absVel, deltaT)
            nextSection, corner = attachedObstacle.checkBounds(
                self, deltaX, deltaY)

            # Check for collisions in this path.
            obstacle = self.world.physics.moveUnit(
                self, corner[0] - self.pos[0], corner[1] - self.pos[1],
                ignoreObstacles=attachedObstacle.subshape)
            if obstacle is not None:
                break
            if nextSection is None:
                obstacle = None
                self.setAttachedObstacle(None)
                break
            if nextSection == attachedObstacle:
                obstacle = None
                break

            self.attachedObstacle = attachedObstacle = nextSection
            nextSection.hitByPlayer(self)
            deltaT -= (
                (oldPos[0] - self.pos[0]) ** 2
                + (oldPos[1] - self.pos[1]) ** 2) ** 0.5 / abs(absVel)
        else:
            log.error(
                'Very very bad thing: motion loop did not terminate after '
                '100 iterations')

        if obstacle:
            if obstacle.jumpable:
                if self.isOnGround() and obstacle.grabbable:
                    # Running into a vertical wall while on the ground.
                    pass
                else:
                    self.setAttachedObstacle(obstacle)
            else:
                self.setAttachedObstacle(None)
            obstacle.hitByPlayer(self)

        self._ignore = None

    def setAttachedObstacle(self, obstacle):
        if obstacle and not obstacle.jumpable:
            log.warning('Cannot set attached obstacle to %s', obstacle)
            return
        self.attachedObstacle = obstacle
        self.motionState = self.getAttachedMotionState(obstacle)

    @staticmethod
    def getAttachedMotionState(obstacle):
        if obstacle is None:
            return 'fall'
        if obstacle.grabbable:
            if obstacle.deltaPt[1] < 0:
                return 'rightwall'
            return 'leftwall'
        return 'ground'

    def calculateJumpMotion(self, absVel):
        '''
        Returns (dx, dy, dt) where (dx, dy) is the trajectory due to the upward
        thrust of jumping, and dt is the time for which the upward thrust does
        not apply.
        '''
        deltaT = self.world.tickPeriod
        deltaY = 0
        deltaX = absVel * deltaT

        # If the player is jumping, calculate how much they jump by.
        if self._jumpTime > 0:
            thrustTime = min(deltaT, self._jumpTime)
            self.yVel = -self.world.physics.playerJumpThrust
            deltaY = thrustTime * self.yVel
            self._jumpTime = self._jumpTime - deltaT

            # Automatically switch off the jumping state if the player
            # has reached maximum time.
            if self._jumpTime <= 0:
                self._jumpTime = 0
            return deltaX, deltaY, deltaT - thrustTime
        return deltaX, deltaY, deltaT

    def getFallTrajectory(self, deltaX, deltaY, fallTime):
        # If player is falling, calculate how far they fall.

        # v = u + at
        vFinal = self.yVel + self.world.physics.playerGravity * fallTime
        if vFinal > self.world.physics.playerMaxFallVel:
            # Hit terminal velocity. Fall has two sections.
            deltaY = (
                deltaY + (
                    self.world.physics.playerMaxFallVel ** 2 - self.yVel ** 2
                ) / (2 * self.world.physics.playerGravity)
                + self.world.physics.playerMaxFallVel * (fallTime - (
                    self.world.physics.playerMaxFallVel - self.yVel
                ) / self.world.physics.playerGravity))

            self.yVel = self.world.physics.playerMaxFallVel
        else:
            # Simple case: s=ut+0.5at**2
            deltaY = (
                deltaY + self.yVel * fallTime + 0.5 *
                self.world.physics.playerGravity * fallTime ** 2)
            self.yVel = vFinal

        return deltaX, deltaY

    def fallThroughAir(self, absVel):
        deltaX, deltaY, fallTime = self.calculateJumpMotion(absVel)
        deltaX, deltaY = self.getFallTrajectory(deltaX, deltaY, fallTime)

        obstacle = self.world.physics.moveUnit(self, deltaX, deltaY)
        if obstacle:
            if obstacle == self._unstickyWall:
                targetPt = obstacle.unstickyWallFinalPosition(
                    self.pos, deltaX, deltaY)
            else:
                targetPt = obstacle.finalPosition(self, deltaX, deltaY)
            if targetPt is not None:
                deltaX = targetPt[0] - self.pos[0]
                deltaY = targetPt[1] - self.pos[1]
                obstacle = self.world.physics.moveUnit(self, deltaX, deltaY)

        if obstacle and obstacle.jumpable:
            self.setAttachedObstacle(obstacle)
        else:
            self.setAttachedObstacle(None)
        if obstacle:
            obstacle.hitByPlayer(self)

        self._ignore = None

    def updateLivingPlayer(self):
        self.processJumpState()

        absVel = self.getAbsVelocity()
        wall = self.isAttachedToWall()

        # Allow falling through fall-through-able obstacles
        if self.isOnPlatform() and self._state['down'] and self.canMove:
            self.dropThroughFloor()
        if self.isOnGround():
            self.moveAlongGround(absVel)
        elif wall:
            if (
                    self._state['down']
                    or (self._state['right'] and wall == 'left')
                    or (self._state['left'] and wall == 'right')
                    or not self.canMove
                    ):
                self.dropOffWall()
        else:
            self.fallThroughAir(absVel)

    def advance(self):
        '''Called by this player's universe when this player should update
        its position. deltaT is the time that's passed since its state was
        current, measured in seconds.'''

        if self.resyncing:
            return

        self.checkUpgrades()
        self.reloadTime = max(0.0, self.reloadTime - self.world.tickPeriod)

        if self.knowsItsDead:
            self.updateGhost()
        elif self.turret:
            self.updateTurret()
        else:
            self.updateLivingPlayer()

        if self.attachedObstacle:
            self.pos = self.attachedObstacle.discretise(Player, self.pos)
        else:
            self.pos = (floor(self.pos[0] + 0.5), floor(self.pos[1] + 0.5))

    def isOnside(self, trosballPosition):
        onLeftOfBall = self.pos[0] < trosballPosition[0]
        onBlueTeam = self.team == self.world.teams[0]
        return onLeftOfBall == onBlueTeam

    def hitShield(self, shooter, deathType):
        '''
        Hits the shield for one point of damage, and returns True if the shield
        absorbed the damage, False if there was no shield.
        '''
        if self.hasProtectiveUpgrade():
            self.upgrade.protections -= 1
            if self.upgrade.protections <= 0:
                self.deleteUpgrade()
                self.onShieldDestroyed(shooter, deathType)
                if self in self.world.players and shooter:
                    shooter.onDestroyedShield(self, deathType)
            return True
        return False

    def zombieHit(self, shooter, shot, deathType):
        '''
        Called when the player has been hit, but the player's client may not
        realise that the player is dead yet.
        '''
        self.zombieHits += 1
        self.onZombieHit(shooter, shot, deathType)

    def properHit(self, shooter, deathType):
        '''
        Called when the player has been hit AND the client is guaranteed to be
        the first one to notice it (because it relates to where the player
        thinks it is rather than where the server thinks the player is).
        '''
        if not self.hitShield(shooter, deathType):
            self.health -= 1
            self.checkDeath(shooter, deathType)

    def noticeZombieHit(self, shooter, deathType):
        '''
        Called when the player notices that it's been hit.
        '''
        assert self.zombieHits >= 0
        self.zombieHits -= 1
        self.properHit(shooter, deathType)

    def killOutright(self, deathType, killer=None, resync=True):
        self.health = 0
        self.checkDeath(killer, deathType)

        # An outright kill cannot be detected by the client until it happens,
        # so the client will have mis-predicted where the player is going and
        # will need to resync.
        if resync:
            self.sendResync(reason='')

    def checkDeath(self, killer, deathType):
        if self.dead:
            if self.hasTrosball():
                self.world.trosballLost(self)
            if self.hasElephant():
                self.world.elephantKill(killer)

        if self.knowsItsDead:
            self.health = 0
            self.zombieHits = 0
            self.respawnGauge = self.world.physics.playerRespawnTotal
            self._jumpTime = 0
            self.pendingUpgrade = None
            self.deleteUpgrade()
            self._setStarsForDeath()
            self.setAttachedObstacle(None)
            if self in self.world.players:
                if killer:
                    killer.onKilled(self, deathType)
                    self.world.onPlayerKill(killer, self, deathType)
                self.world.playerHasNoticedDeath(self, killer)
            self.onDied(killer, deathType)

    def _setStarsForDeath(self):
        self.starsAtLastDeath = self.stars
        self.stars = self.world.PLAYER_RESET_STARS
        self.onStarsChanged()

    def respawn(self, zone=None):
        if self.resyncing:
            return
        self.setAttachedObstacle(None)
        self.health = self.world.physics.playerRespawnHealth
        self.zombieHits = 0
        self.invulnerableUntil = (
            self.world.timer.getElapsedTime() + RESPAWN_CAMP_TIME)
        self.respawnGauge = 0
        self.teleportToZoneCentre(zone)
        self.world.onPlayerRespawn(self)
        self.onRespawned()

    def teleportToZoneCentre(self, zone=None):
        if zone is None:
            zone = self.getZone()
        self.setPos(zone.defn.pos, 'f')

    def deleteUpgrade(self, abandoned=False):
        '''Deletes the current upgrade object from this player'''
        if self.resyncing:
            return
        upgrade = self.upgrade
        if upgrade:
            upgrade.delete()
            if abandoned:
                self.onAbandonUpgrade(upgrade)
            self.onUpgradeGone(upgrade)

    def incrementStars(self):
        if self.stars < MAX_STARS:
            self.stars += 1
            self.onStarsChanged()

    def getShotType(self):
        '''
        Returns one of the following:
            None - if the player cannot currently shoot
            'N' - if the player can shoot normal shots
            'R' - if the player can shoot ricochet shots
            'T' - if the player can shoot turret shots
        '''
        if self.dead or self.phaseshift or not self.canMove:
            return None

        # While on a vertical wall, one canst not fire
        if self.reloadTime > 0 or self.isAttachedToWall():
            return None

        if self.hasRicochet:
            return Shot.RICOCHET
        if self.turret:
            if self.turretOverHeated:
                return None
            return Shot.TURRET
        return Shot.NORMAL

    def setUser(self, user):
        if not self.bot:
            self.user = user

    def isElephantOwner(self):
        return self.user is not None and self.user.isElephantOwner()

    def hasElephant(self):
        playerWithElephant = self.world.playerWithElephant
        return playerWithElephant and playerWithElephant.id == self.id

    def canShoot(self):
        return self.getShotType() is not None

    def sendResync(
            self, reason='Your computer was out of sync with the server!'):
        args = self.getPlayerUpdateArgs(resync=True)
        globalMsg, resyncMsg = PlayerUpdateMsg(*args), ResyncPlayerMsg(*args)
        if self.resyncing and self.checkResyncAcknowledgement(resyncMsg):
            # No need to send the second resync, the player is still in the
            # same state.
            return
        if not self.world.isServer:
            return
        self.world.sendServerCommand(globalMsg)
        if self.agent:
            self.agent.messageToAgent(resyncMsg)
            if reason:
                self.agent.messageToAgent(
                    ChatFromServerMsg(error=True, text=reason))
        self.resyncBegun()

    def resyncBegun(self):
        self.resyncing = True
        self.resyncExpiry = self.world.lastTickId + int(
            MAX_RESYNC_TIME / self.world.tickPeriod)

    def getPlayerUpdateArgs(self, resync=None):
        attached = self.motionState[0]

        if resync is None:
            resync = self.resyncing

        health = max(self.health - self.zombieHits, 0)

        return (
            self.id, self.pos[0], self.pos[1], self.yVel, self.angleFacing,
            self.ghostThrust, self._jumpTime, self.reloadTime,
            self.respawnGauge, attached, self.stars, health, resync,
            self._state['left'], self._state['right'], self._state['jump'],
            self._state['down'])

    def applyPlayerUpdate(self, msg):
        self._state['left'] = msg.leftKey
        self._state['right'] = msg.rightKey
        self._state['jump'] = msg.jumpKey
        self._state['down'] = msg.downKey

        self.yVel = msg.yVel
        self.lookAt(msg.angle, msg.ghostThrust)
        self.setPos((msg.xPos, msg.yPos), msg.attached)
        self._jumpTime = msg.jumpTime
        self.reloadTime = msg.gunReload
        self.respawnGauge = msg.respawn
        oldStars = self.stars
        self.stars = msg.stars
        self.health = msg.health
        self.zombieHits = 0
        if msg.resync:
            self.resyncBegun()
        else:
            self.resyncing = False
        if oldStars != self.stars:
            self.onStarsChanged()

    def buildResyncAcknowledgement(self):
        return ResyncAcknowledgedMsg(
            self.world.lastTickId, self.pos[0], self.pos[1],
            self.yVel, self.angleFacing, self.ghostThrust)

    def checkResyncAcknowledgement(self, msg):
        '''
        Checks whether the player position etc. matches the position encoded in
        the ResyncPlayerMsg or ResyncAcknowledgedMsg.
        '''
        return (
            isNear(self.pos[0], msg.xPos)
            and isNear(self.pos[1], msg.yPos)
            and isNear(self.yVel, msg.yVel)
            and isNear(self.angleFacing, msg.angle)
            and isNear(self.ghostThrust, msg.ghostThrust))

    def weaponDischarged(self):
        '''
        Updates the player's reload time because the player has just fired a
        shot.
        '''
        world = self.world
        zone = self.getZone()
        if self.turret:
            reloadTime = world.physics.playerTurretReloadTime
            self.turretHeat += world.physics.playerShotHeat
            if self.turretHeat > world.physics.playerTurretHeatCapacity:
                self.turretOverHeated = True
        elif self.machineGunner:
            self.mgBulletsRemaining -= 1
            if self.mgBulletsRemaining > 0:
                reloadTime = world.physics.playerMachineGunFireRate
            elif self.mgBulletsRemaining == 0:
                reloadTime = world.physics.playerMachineGunReloadTime
            else:
                self.mgBulletsRemaining = 15
                reloadTime = 0
        elif self.shoxwave:
            reloadTime = world.physics.playerShoxwaveReloadTime
        elif self.team is None:
            reloadTime = world.physics.playerNeutralReloadTime
        elif world.state.isTrosball():
            if abs(self.pos[0] - world.getTrosballPosition()[0]) < 1e-5:
                reloadTime = world.physics.playerNeutralReloadTime
            # If self is on blue, and on the left of the trosball; or
            # completely vice versa
            elif self.isOnside(world.getTrosballPosition()):
                reloadTime = world.physics.playerOwnDarkReloadTime
            else:
                reloadTime = world.physics.playerEnemyDarkReloadTime
        elif zone and zone.owner == self.team and zone.dark:
            reloadTime = world.physics.playerOwnDarkReloadTime
        elif zone and not zone.dark:
            reloadTime = world.physics.playerNeutralReloadTime
        else:
            reloadTime = world.physics.playerEnemyDarkReloadTime
        self.reloadTime = self.reloadFrom = reloadTime

    def createShot(self, shotId=None, shotClass=Shot):
        '''
        Factory function for building a Shot object.
        '''
        # Shots take some fraction of the player's current velocity, but we
        # still want the shots to always go directly towards where the user
        # clicked, so we take the component of player velocity in the direction
        # of the shot.
        f = self.world.physics.fractionOfPlayerVelocityImpartedToShots
        xVel = (self.pos[0] - self.oldPos[0]) / self.world.tickPeriod
        yVel = (self.pos[1] - self.oldPos[1]) / self.world.tickPeriod
        shotVel = self.world.physics.shotSpeed + f * (
            xVel * sin(self.angleFacing) - yVel * cos(self.angleFacing))

        velocity = (
            shotVel * sin(self.angleFacing),
            -shotVel * cos(self.angleFacing),
        )

        kind = self.getShotType()
        lifetime = self.world.physics.shotLifetime

        team = self.team
        shot = shotClass(
            self.world, shotId, team, self, self.pos, velocity, kind, lifetime)
        return shot



import copy
import cPickle
import heapq
import logging
from math import pi
import os
import weakref
import zipfile

from twisted.internet import defer, reactor, threads

from trosnoth import data
from trosnoth.model.mapLayout import LayoutDatabase
from trosnoth.model.player import Player
from trosnoth.model.universe import Universe

log = logging.getLogger(__name__)


DB_FILENAME = data.getPath(data, 'pathfinding.db')

NORTH = 'n'
SOUTH = 's'
EAST = 'e'
WEST = 'w'
ORB = 'o'

OPPOSITE_DIRECTION = {
    NORTH: SOUTH,
    SOUTH: NORTH,
    EAST: WEST,
    WEST: EAST,
}

ADJACENT_KINDS = {
    'topBlocked': {
        NORTH: [],
        SOUTH: ['btmBlocked', 'btmOpen'],
        EAST: ['bck'],
        WEST: ['fwd'],
    },
    'topOpen': {
        NORTH: ['btmBlocked', 'btmOpen'],
        SOUTH: ['btmBlocked', 'btmOpen'],
        EAST: ['bck'],
        WEST: ['fwd'],
    },
    'btmBlocked': {
        NORTH: ['topBlocked', 'topOpen'],
        SOUTH: [],
        EAST: ['fwd'],
        WEST: ['bck'],
    },
    'btmOpen': {
        NORTH: ['topBlocked', 'topOpen'],
        SOUTH: ['topBlocked', 'topOpen'],
        EAST: ['fwd'],
        WEST: ['bck'],
    },
    'fwd': {
        NORTH: ['bck'],
        SOUTH: ['bck'],
        EAST: ['topBlocked', 'topOpen'],
        WEST: ['btmBlocked', 'btmOpen'],
    },
    'bck': {
        NORTH: ['fwd'],
        SOUTH: ['fwd'],
        EAST: ['btmBlocked', 'btmOpen'],
        WEST: ['topBlocked', 'topOpen'],
    },
}


class ActionKindCollection(object):
    def __init__(self):
        self.store = {}

    def register(self, cls):
        key = cls.__name__
        if key in self.store:
            raise KeyError('Another %s already registered', key)
        self.store[key] = cls
        return cls

    def getByString(self, key):
        return self.store[key]

    def actionKindToString(self, actionKind):
        key = actionKind.__name__
        assert self.store[key] == actionKind
        return key

actionKindClasses = ActionKindCollection()


class PathFindingAction(object):
    '''
    Represents a possible action for player path-finding.
    '''

    def prepNextStep(self, player):
        '''
        Returns (done, stateChangesfaceRight) where done is a boolean
        indicating whether or not this action is complete, stateChanges is a
        list of (key, value) pairs indicating necessary changes to the player's
        getState() dict, and faceRight is whether or not the player should be
        facing right for this step.
        '''
        raise NotImplementedError(
            '%s.prepNextStep()' % (self.__class__.__name__,))

    @classmethod
    def buildStateChanges(cls, player, desiredState):
        result = []
        for k, v in player._state.iteritems():
            desired = desiredState.get(k, False)
            if v != desired:
                result.append((k, desired))
        return result


class SingleAction(PathFindingAction):

    def __init__(self):
        self.first = True

    @classmethod
    def simulate(cls, player, alsoAllowBlockDefs=()):
        '''
        Applies this action type to the given player object until it is
        complete or it leaves the current map block.
        '''
        action = cls()

        orbTagCost = None

        startMapBlock = player.getMapBlock()
        cost = 0
        while True:
            done, stateChanges, faceRight = action.prepNextStep(player)
            if done:
                break
            for key, value in stateChanges:
                player.updateState(key, value)
            if faceRight and not player.isFacingRight():
                player.lookAt(pi / 2)
            elif (not faceRight) and player.isFacingRight():
                player.lookAt(-pi / 2)
            cost += 1

            player.advance()
            zone = player.getZone()
            if (
                    orbTagCost is None and zone
                    and zone.playerIsWithinTaggingDistance(player)):
                orbTagCost = cost
            playerBlock = player.getMapBlock()
            if playerBlock != startMapBlock:
                if playerBlock.defn not in alsoAllowBlockDefs:
                    break

            if cost >= 300:
                log.warning(
                    'Action %s looks like it will go forever' % (cls,))
                return None, None

        return cost, orbTagCost


@actionKindClasses.register
class FallDown(SingleAction):
    def prepNextStep(self, player):
        stateChanges = self.buildStateChanges(player, {})
        done = not self.first and (
            player.isOnGround() or player.isAttachedToWall())
        self.first = False
        return done, stateChanges, player.isFacingRight()


@actionKindClasses.register
class FallLeft(SingleAction):
    def prepNextStep(self, player):
        stateChanges = self.buildStateChanges(player, {'left': True})
        done = not self.first and (
            player.isOnGround() or player.isAttachedToWall())
        self.first = False
        return done, stateChanges, False


@actionKindClasses.register
class FallRight(SingleAction):
    def prepNextStep(self, player):
        stateChanges = self.buildStateChanges(player, {'right': True})
        done = not self.first and (
            player.isOnGround() or player.isAttachedToWall())
        self.first = False
        return done, stateChanges, True


@actionKindClasses.register
class Drop(SingleAction):
    def prepNextStep(self, player):
        stateChanges = self.buildStateChanges(player, {'down': True})
        done = not self.first
        self.first = False
        return done, stateChanges, player.isFacingRight()


@actionKindClasses.register
class JumpUp(SingleAction):
    def prepNextStep(self, player):
        if self.first:
            stateChanges = self.buildStateChanges(player, {}) + [
                ('jump', True)]
            done = False
        elif player._jumpTime > 0:
            stateChanges = self.buildStateChanges(player, {'jump': True})
            done = False
        else:
            stateChanges = self.buildStateChanges(player, {})
            done = player.yVel >= 0
        self.first = False
        return done, stateChanges, player.isFacingRight()


@actionKindClasses.register
class JumpLeft(SingleAction):
    def prepNextStep(self, player):
        if self.first:
            stateChanges = self.buildStateChanges(player, {'left': True}) + [
                ('jump', True)]
            done = False
        elif player._jumpTime > 0:
            stateChanges = self.buildStateChanges(
                player, {'left': True, 'jump': True})
            done = False
        else:
            stateChanges = self.buildStateChanges(player, {'left': True})
            done = player.yVel >= 0
        self.first = False
        return done, stateChanges, False


@actionKindClasses.register
class JumpRight(SingleAction):
    def prepNextStep(self, player):
        if self.first:
            stateChanges = self.buildStateChanges(player, {'right': True}) + [
                ('jump', True)]
            done = False
        elif player._jumpTime > 0:
            stateChanges = self.buildStateChanges(
                player, {'right': True, 'jump': True})
            done = False
        else:
            stateChanges = self.buildStateChanges(player, {'right': True})
            done = player.yVel >= 0
        self.first = False
        return done, stateChanges, True


@actionKindClasses.register
class MoveLeft(SingleAction):
    def prepNextStep(self, player):
        if self.first:
            stateChanges = self.buildStateChanges(player, {'left': True})
            done = False
        else:
            stateChanges = []
            done = True
        self.first = False
        return done, stateChanges, False


@actionKindClasses.register
class MoveRight(SingleAction):
    def prepNextStep(self, player):
        if self.first:
            stateChanges = self.buildStateChanges(player, {'right': True})
            done = False
        else:
            stateChanges = []
            done = True
        self.first = False
        return done, stateChanges, True


@actionKindClasses.register
class SlowStepLeft(SingleAction):
    def prepNextStep(self, player):
        if self.first:
            stateChanges = self.buildStateChanges(player, {'left': True})
            done = False
        else:
            stateChanges = []
            done = True
        self.first = False
        return done, stateChanges, True


@actionKindClasses.register
class SlowStepRight(SingleAction):
    def prepNextStep(self, player):
        if self.first:
            stateChanges = self.buildStateChanges(player, {'right': True})
            done = False
        else:
            stateChanges = []
            done = True
        self.first = False
        return done, stateChanges, False


class MultiActionPath(PathFindingAction):
    def __init__(self, actions):
        super(MultiActionPath, self).__init__()
        self.actions = [a() for a in actions]

    def prepNextStep(self, player):
        if not self.actions:
            return True, [], player.isFacingRight()
        while self.actions:
            done, stateChanges, faceRight = self.actions[0].prepNextStep(
                player)
            if done:
                self.actions.pop(0)
            else:
                break
        return (len(self.actions) == 0), stateChanges, faceRight


class RouteToOrb(PathFindingAction):
    def __init__(self, pathFinder, zone):
        super(RouteToOrb, self).__init__()
        self.pathFinder = pathFinder
        self.zone = zone
        self.currentActions = []
        self.steps = None  # Not yet calculated

    def prepNextStep(self, player):
        if self.steps is None:
            if not player.attachedObstacle:
                # Just fall. When we land we can work out what to do.
                return False, [], player.isFacingRight()
            self.calculatePath(player)

        while True:
            if not self.steps:
                # Finished
                return True, [], player.isFacingRight()

            blockDef, scoreKey = self.steps[0]
            if (
                    scoreKey == ORB and
                    player.getZone().playerIsWithinTaggingDistance(player)):
                self.steps = []
                return True, [], player.isFacingRight()

            while self.currentActions:
                done, changes, faceRight = self.currentActions[0].prepNextStep(
                    player)
                if done:
                    self.currentActions.pop(0)
                else:
                    return False, changes, faceRight

            playerMapBlock = player.getMapBlock()
            if playerMapBlock.defn != blockDef:
                self.steps.pop(0)
                if self.steps and playerMapBlock.defn != self.steps[0][0]:
                    log.warning('Ended up in wrong map block!')
                    self.steps = []
                if not self.steps:
                    return True, [], False
                blockDef, scoreKey = self.steps[0]

            blockPaths = self.pathFinder.blocks[blockDef]
            self.currentActions = [
                a() for a in blockPaths.getActions(player, scoreKey)]

            if not self.currentActions:
                # Error case, act like we're done.
                self.steps = []

    def calculatePath(self, player):
        '''
        Use Dijkstra's algorithm to calculate the shortest path through the map
        blocks based on the approximate traversal costs of the map blocks.
        '''
        self.steps = []

        block = player.getMapBlock()
        blockPaths = self.pathFinder.blocks[block.defn]
        playerNode = blockPaths.getNode(player)
        if player.getZone() == self.zone and ORB in playerNode['scores']:
            self.steps.append((block.defn, ORB))
            return

        targets = {}
        for direction in [NORTH, SOUTH, EAST, WEST]:
            if direction in playerNode['scores']:
                targets[blockPaths.getMapBlockNode(direction)] = (
                    playerNode['scores'][direction])

        heap = []
        seen = set()
        finished = object()

        for node in self.pathFinder.getOrbNodes(self.zone):
            heapq.heappush(heap, (0, None, node))

        while heap:
            cost, path, node = heapq.heappop(heap)
            if node in seen:
                continue
            seen.add(node)
            if node is finished:
                break
            if node in targets:
                stepCost = targets[node]
                heapq.heappush(heap, (cost + stepCost, (node, path), finished))
                continue

            for child, stepCost in node.getIncomingPaths():
                heapq.heappush(heap, (cost + stepCost, (node, path), child))

        else:
            # No path found!
            log.warning('No path to orb could be found!')
            return

        while path is not None:
            node, path = path
            self.steps.append((node.blockDef, node.scoreKey))


class PathFindingStore(object):
    def __init__(self):
        self.blocks = {}
        self.combos = {}

    def build(self):
        log.info('Finding allowed nodes and edges...')

        layoutDb = LayoutDatabase()
        ds = layoutDb.datastores[0]
        for (kind, blocked), layouts in ds.layouts.iteritems():
            for layout in layouts:
                log.info('%s', os.path.basename(layout.filename))
                self.addBlock(layout, kind, blocked, layoutDb=layoutDb)

        log.info('Calculating walk scores within blocks...')
        for blockPaths in self.blocks.itervalues():
            blockPaths.calculateDefaultWalkScores()

        log.info('Calculating walk scores between blocks...')
        self.buildBlockCombinations(layoutDb=layoutDb)

        self.save()

    def addBlock(self, blockLayout, kind, blocked, layoutDb=None):
        '''
        Adds the given map block layout to the store and performs the initial
        pass of calculations.
        '''
        if blockLayout.key in self.blocks:
            raise KeyError('%s already in store' % (blockLayout.filename,))
        paths = PreCalculatedMapBlockPaths(blockLayout, kind, blocked)
        self.blocks[blockLayout.key] = paths
        paths.initialiseNodesAndEdges(layoutDb)

    def buildBlockCombinations(self, layoutDb=None):
        '''
        Goes through combinations of blocks which can fit together and
        calculates the costs of walking between the blocks.
        '''
        byKind = {}
        for blockPaths in self.blocks.itervalues():
            key = blockPaths.getKindKey()
            byKind.setdefault(key, []).append(blockPaths)

        for startKind, blockPaths in byKind.iteritems():
            for startBlockPaths in blockPaths:
                for direction, allowedKinds in (
                        ADJACENT_KINDS[startKind].iteritems()):
                    for endKind in allowedKinds:
                        for endBlockPaths in byKind.get(endKind, []):
                            self.addCombo(
                                startBlockPaths, endBlockPaths, direction,
                                layoutDb=layoutDb)

    def addCombo(
            self, startBlockPaths, endBlockPaths, direction, layoutDb=None):
        combo = PreCalculatedMapBlockCombination(
            startBlockPaths, endBlockPaths, direction)
        key = startBlockPaths.layout.key, endBlockPaths.layout.key, direction
        self.combos[key] = combo
        combo.calculateConnections(layoutDb)
        combo.calculateWalkScores()

    def save(self):
        with zipfile.ZipFile(DB_FILENAME, 'w') as z:
            for block in self.blocks.itervalues():
                block.save(z)
            for combo in self.combos.itervalues():
                combo.save(z)


class RunTimePathFinder(object):

    def __init__(self, mapLayout):
        self.layout = mapLayout
        self.blocks = {}

    @defer.inlineCallbacks
    def loadData(self):
        '''
        This method loads from the path finding database, and yields every so
        often so that it can be used to load progressively. This is
        deliberately NOT @inlineCallbacks, in order that the caller can have
        more fine control of how often it loads the next step.
        '''
        blockCache = {}
        comboCache = {}

        with zipfile.ZipFile(DB_FILENAME, 'r') as z:
            for row in self.layout.blocks:
                for blockDef in row:
                    if blockDef.layout is None:
                        continue
                    runTimePaths = yield (
                        PreCalculatedMapBlockPaths.applyToBlock(
                            blockDef, z, blockCache))
                    if runTimePaths:
                        self.blocks[blockDef] = runTimePaths

                    d = defer.Deferred()
                    reactor.callLater(0, d.callback, None)
                    yield d

            for j, row1 in enumerate(self.layout.blocks[:-1]):
                row2 = self.layout.blocks[j + 1]
                for i, blockDef1 in enumerate(row1):
                    blockDef2 = row2[i]
                    if blockDef1.layout is None or blockDef2.layout is None:
                        continue
                    if blockDef1.blocked:
                        continue
                    yield self.blocks[blockDef1].updateForCombination(
                        SOUTH, self.blocks[blockDef2], z, comboCache)
                    yield self.blocks[blockDef2].updateForCombination(
                        NORTH, self.blocks[blockDef1], z, comboCache)

                    d = defer.Deferred()
                    reactor.callLater(0, d.callback, None)
                    yield d

            for row in self.layout.blocks:
                for i, blockDef1 in enumerate(row[:-1]):
                    blockDef2 = row[i + 1]
                    if blockDef1.layout is None or blockDef2.layout is None:
                        continue
                    yield self.blocks[blockDef1].updateForCombination(
                        EAST, self.blocks[blockDef2], z, comboCache)
                    yield self.blocks[blockDef2].updateForCombination(
                        WEST, self.blocks[blockDef1], z, comboCache)

                    d = defer.Deferred()
                    reactor.callLater(0, d.callback, None)
                    yield d

    def getOrbAction(self, player):
        block = player.getMapBlock()
        if block.defn not in self.blocks:
            return None
        return MultiActionPath(self.blocks[block.defn].getActions(player, ORB))

    def getExitAction(self, player, direction):
        block = player.getMapBlock()
        if block.defn not in self.blocks:
            return None
        return MultiActionPath(
            self.blocks[block.defn].getActions(player, direction))

    def getAllActions(self, player):
        block = player.getMapBlock()
        if block.defn not in self.blocks:
            return None
        return self.blocks[block.defn].getAllActions(player)

    def getRouteToOrb(self, zone):
        return RouteToOrb(self, zone)

    def getOrbNodes(self, zone):
        x, y = zone.defn.pos
        j, i = self.layout.getMapBlockIndices(x, y - 5)
        blockDef1 = self.layout.blocks[j][i]
        blockDef2 = self.layout.blocks[j + 1][i]
        yield self.blocks[blockDef1].getMapBlockNode(ORB)
        yield self.blocks[blockDef2].getMapBlockNode(ORB)


class PreCalculatedMapBlockPaths(object):
    def __init__(self, blockLayout, kind, blocked):
        self.kind = kind
        self.blocked = blocked
        self.layout = blockLayout
        self.nodes = {}
        self.outbound = {
            NORTH: PreCalculatedNode(self),
            SOUTH: PreCalculatedNode(self),
            EAST: PreCalculatedNode(self),
            WEST: PreCalculatedNode(self),
        }
        self.orbEdges = set()

    def save(self, z):
        arcName = 'blocks/' + '/'.join(str(i) for i in self.layout.key)
        contents = cPickle.dumps({
            'nodes': [node.serialise() for node in self.nodes.itervalues()],
        }, protocol=cPickle.HIGHEST_PROTOCOL)
        z.writestr(arcName, contents, zipfile.ZIP_DEFLATED)

    @staticmethod
    @defer.inlineCallbacks
    def applyToBlock(blockDef, z, blockCache):
        arcName = 'blocks/' + '/'.join(str(i) for i in blockDef.layout.key)
        if arcName not in blockCache:
            yield threads.deferToThread(
                PreCalculatedMapBlockPaths._loadPathData,
                blockDef, z, arcName, blockCache)

        data = blockCache[arcName]
        defer.returnValue(RunTimeMapBlockPaths(blockDef, data))

    @staticmethod
    def _loadPathData(blockDef, z, arcName, blockCache):
        try:
            f = z.open(arcName)
        except KeyError:
            log.warning(
                'No path finding data found for %s',
                os.path.basename(blockDef.layout.filename)
                + (' (reverse)' if blockDef.layout.reversed else ''),
            )
            blockCache[arcName] = None
            return

        with f:
            blockCache[arcName] = cPickle.load(f)

    def getWorldAndBlockDef(self, layoutDb=None):
        '''
        Creates and returns a dummy world and the block definition
        corresponding to this map block.
        '''
        if layoutDb is None:
            layoutDb = LayoutDatabase()

        world = Universe(layoutDb, 1, 1)

        if self.layout.kind in ('fwd', 'bck'):
            i, j = 2, 1
        elif self.layout.kind == 'top':
            i, j = 1, 1
        else:
            i, j = 1, 2

        block = world.map.layout.blocks[j][i]
        self.layout.applyTo(block)

        return world, block

    def initialiseNodesAndEdges(self, layoutDb=None):
        world, block = self.getWorldAndBlockDef(layoutDb)

        for obstacle in block.obstacles + block.ledges:
            if not obstacle.jumpable:
                continue
            minI, maxI = obstacle.getPositionIndexBounds(Player)
            for i in xrange(minI, maxI + 1):
                node = PreCalculatedNode(self, block, obstacle, i)
                self.nodes[node.obstacleId, i] = node

        for node in self.nodes.itervalues():
            node.expandEdges(world, block)

        log.info(
            '  -> Nodes: %r  Edges: %r  Exits: %r  Orb caps: %r',
            len(self.nodes),
            sum(len(n.outbound) for n in self.nodes.itervalues()),
            sum(len(n.inbound) for n in self.outbound.itervalues()),
            len(self.orbEdges),
        )

    def nodeFromPlayerState(self, player):
        '''
        Finds and returns the node that the given player state is associated
        with.
        '''
        assert player.attachedObstacle.jumpable
        obstacleId = player.attachedObstacle.obstacleId
        posIndex = player.attachedObstacle.getPositionIndex(Player, player.pos)
        return self.nodes[obstacleId, posIndex]

    def getExitNode(self, startBlock, endBlock):
        x0, y0 = startBlock.pos
        x1, y1 = endBlock.pos
        if x1 == x0 and y1 < y0:
            return self.outbound[NORTH]
        if x1 == x0 and y1 > y0:
            return self.outbound[SOUTH]
        if x1 > x0 and y1 == y0:
            return self.outbound[EAST]
        if x1 < x0 and y1 == y0:
            return self.outbound[WEST]
        return None

    def calculateDefaultWalkScores(self):
        '''
        Calculates the minimum distance from each node to any possible way of
        exiting the zone in each direction.
        '''
        for d in [NORTH, SOUTH, EAST, WEST]:
            self.calculateWalkScores(self.outbound[d].inbound, d)
        self.calculateWalkScores(self.orbEdges, ORB, orbCosts=True)

    def calculateWalkScores(self, targetEdges, scoreKey, orbCosts=False):
        '''
        Uses Dijkstra's algorithm to calculate the minimum distance from each
        node in this block to complete any of the given target edges. Stores
        these minimum values using the given scoreKey. If orbCosts is True,
        uses the orb cost of the given edges rather than the actual cost.
        '''
        heap = []
        for edge in targetEdges:
            if orbCosts:
                cost = edge.orbTagCost
            else:
                cost = edge.cost
            heapq.heappush(heap, (cost, edge.startNode))

        seen = set()
        while heap:
            cost, node = heapq.heappop(heap)
            if node in seen:
                continue
            seen.add(node)
            node.setWalkScore(scoreKey, cost)
            for edge in node.inbound:
                if edge.startNode in seen:
                    continue
                heapq.heappush(heap, (cost + edge.cost, edge.startNode))
        log.info(
            '%s -> %s: Visited %r nodes (/%r)',
            os.path.basename(self.layout.filename), scoreKey,
            len(seen), len(self.nodes))

    def getKindKey(self):
        if self.kind in ('top', 'btm'):
            return '%s%s' % (self.kind, 'Blocked' if self.blocked else 'Open')
        return self.kind


class RunTimeMapBlockPaths(object):
    def __init__(self, blockDef, blockData):
        self.blockDef = blockDef

        # Calling copy.deepcopy() on the whole node data takes a long time, so
        # select the ones we're going to be mutating and copy them only.
        self.nodes = {
            data['key']: {
                'edges': data['edges'],     # does not need copying
                'targets': copy.deepcopy(data['targets']),
                'scores': copy.deepcopy(data['scores']),
            } for data in blockData['nodes']
        }

        self.mapBlockNodes = {
            k: RunTimeMapBlockNode(blockDef, k)
            for k in [NORTH, SOUTH, EAST, WEST, ORB]}

    @defer.inlineCallbacks
    def updateForCombination(self, direction, other, z, cache):
        '''
        Called during loading, to update this block's node data with
        information based on what map blocks are adjacent to this one.
        Directionality: if you travel in direction from self, you will get to
        other.
        '''
        # Load the data for the combination
        data = yield PreCalculatedMapBlockCombination.loadData(
            self.blockDef, other.blockDef, direction, z, cache)

        if data is None:
            return

        for nodeKey, cost in data['scores'].iteritems():
            if cost is None:
                del self.nodes[nodeKey]['scores'][direction]
            else:
                self.nodes[nodeKey]['scores'][direction] = cost

        for edgeData in data['edges']:
            nodeKey = edgeData['source']
            self.nodes[nodeKey]['targets'][direction] = {
                'cost': edgeData['cost'],
                'actions': edgeData['actions'],
            }

        # Link the high-level path-finding nodes
        if not data['transitions']:
            log.error(
                'No transition data for %s -> %s: %s',
                os.path.basename(self.blockDef.layout.filename),
                direction,
                os.path.basename(other.blockDef.layout.filename))
        for outDir, cost in data['transitions'].iteritems():
            other.mapBlockNodes[outDir].addInwardEdge(
                self.mapBlockNodes[direction], cost)

    def getNode(self, player):
        obstacleId = player.attachedObstacle.obstacleId
        posIndex = player.attachedObstacle.getPositionIndex(Player, player.pos)
        return self.nodes[obstacleId, posIndex]

    def getBestEdge(self, node, scoreKey):
        if scoreKey in node['targets']:
            return node['targets'][scoreKey]

        def endNodeScore(edge):
            endNode = self.nodes[edge['target']]
            return endNode['scores'][scoreKey]

        def edgeScore(edge):
            return endNodeScore(edge) + edge['cost']

        edges = [
            e for e in node['edges']
            if scoreKey in self.nodes[e['target']]['scores']]
        if not edges:
            return None

        best = min(edges, key=edgeScore)
        if endNodeScore(best) >= node['scores'].get(scoreKey, 10000):
            # Even the best action does not get us closer (should never happen)
            log.warning('Stuck in local minimum (should be impossible)')
            log.warning('Perhaps pathfinding DB needs regenerating.')
            log.warning(
                '(in %s, going to %r)',
                os.path.basename(self.blockDef.layout.filename),
                scoreKey)
            return None

        return best

    def getActions(self, player, scoreKey):
        if not player.attachedObstacle:
            return [Drop]
        edge = self.getBestEdge(self.getNode(player), scoreKey)
        if edge is None:
            return []
        return [actionKindClasses.getByString(a) for a in edge['actions']]

    def getAllActions(self, player):
        if not player.attachedObstacle:
            return [MultiActionPath([Drop])]
        return [
            MultiActionPath([
                actionKindClasses.getByString(a) for a in edge['actions']])
            for edge in self.getNode(player)['edges']
        ]

    def getMapBlockNode(self, scoreKey):
        return self.mapBlockNodes[scoreKey]


class RunTimeMapBlockNode(object):
    '''
    Used to calculate the fastest path through a map at run-time. Represents
    exiting the given block by following the given score key.
    '''

    def __init__(self, blockDef, entryDirection):
        self.blockDef = blockDef
        self.scoreKey = entryDirection
        self.inwardEdges = []

    def __str__(self):
        return '(%s from %s)' % (self.scoreKey, self.blockDef)

    def __repr__(self):
        return 'RunTimeMapBlockNode(%r, %r)' % (
            self.blockDef, self.entryDirection)

    def addInwardEdge(self, node, cost):
        '''
        Should be called only during setup.
        '''
        ownKey = self.scoreKey
        if ownKey == ORB:
            ownKey = NORTH if self.blockDef.kind == 'btm' else SOUTH
        if node.scoreKey == OPPOSITE_DIRECTION[ownKey]:
            # Don't bother with edges that loop forever
            return
        self.inwardEdges.append((node, cost))

    def getIncomingPaths(self):
        '''
        Yields (node, cost) for each RunTimeMapBlockNode that can traverse to
        this node.
        '''
        return iter(self.inwardEdges)


class PreCalculatedMapBlockCombination(object):
    def __init__(self, start, end, direction):
        self.start = start
        self.end = end
        self.direction = direction
        self.edges = {}     # start -> edge
        self.walkScores = {}
        # Note: transitionCosts is the average cost of transitioning through
        # the *end* block given that you came from this direction.
        self.transitionCosts = {}

    def save(self, z):
        arcName = 'combos/%s/%s/%s' % (
            '/'.join(str(i) for i in self.start.layout.key),
            self.direction,
            '/'.join(str(i) for i in self.end.layout.key))

        contents = cPickle.dumps({
            'scores': {
                node.serialiseKey(): cost
                for node, cost in self.walkScores.iteritems()
            },
            'edges': [
                edge.serialiseTargetEdge(self.direction, includeSource=True)
                for edge in self.edges.itervalues()],
            'transitions': self.transitionCosts,
        }, protocol=cPickle.HIGHEST_PROTOCOL)
        z.writestr(arcName, contents, zipfile.ZIP_DEFLATED)

    @staticmethod
    @defer.inlineCallbacks
    def loadData(blockDef1, blockDef2, direction, z, cache):
        arcName = 'combos/%s/%s/%s' % (
            '/'.join(str(i) for i in blockDef1.layout.key),
            direction,
            '/'.join(str(i) for i in blockDef2.layout.key))

        if arcName not in cache:
            yield threads.deferToThread(
                PreCalculatedMapBlockCombination._loadCombinationData,
                blockDef1, blockDef2, direction, z, arcName, cache)

        defer.returnValue(cache[arcName])

    @staticmethod
    def _loadCombinationData(
            blockDef1, blockDef2, direction, z, arcName, cache):
        try:
            f = z.open(arcName)
        except KeyError:
            log.warning(
                'No path finding data found for %s -> %s: %s',
                os.path.basename(blockDef1.layout.filename)
                + (' (reverse)' if blockDef1.layout.reversed else ''),
                direction,
                os.path.basename(blockDef2.layout.filename)
                + (' (reverse)' if blockDef2.layout.reversed else ''),
            )
            log.warning(arcName)
            cache[arcName] = None
            defer.returnValue(None)

        with f:
            cache[arcName] = cPickle.load(f)

    def getWorldAndBlockDefs(self, layoutDb=None):
        world, startBlockDef = self.start.getWorldAndBlockDef(layoutDb)

        if self.start.layout.kind in ('fwd', 'bck'):
            i, j = 2, 1
        elif self.start.layout.kind == 'top':
            i, j = 1, 1
        else:
            i, j = 1, 2

        if self.direction == NORTH:
            j -= 1
        elif self.direction == SOUTH:
            j += 1
        elif self.direction == EAST:
            i += 1
        else:
            assert self.direction == WEST
            i -= 1

        endBlockDef = world.map.layout.blocks[j][i]
        self.end.layout.applyTo(endBlockDef)

        return world, startBlockDef, endBlockDef

    def calculateConnections(self, layoutDb=None):
        '''
        Calculates all edges leaving the start block and ending up in an
        allowable place in the end block.
        '''
        log.info(
            '%s -> %s: %s',
            os.path.basename(self.start.layout.filename) + (
                ' (reverse)' if self.start.layout.reversed else ''),
            self.direction,
            os.path.basename(self.end.layout.filename) + (
                ' (reverse)' if self.end.layout.reversed else ''),
        )

        world, startBlockDef, endBlockDef = self.getWorldAndBlockDefs(layoutDb)

        partitions = {}
        edges = set()

        for edge in self.start.outbound[self.direction].inbound:
            startPlayer = edge.startNode.makePlayerState(world, startBlockDef)

            options = [(edge.actions, startPlayer)]
            while options:
                steps, player = options.pop(0)
                lastAction = steps[-1]
                cost, orbTagCost = lastAction.simulate(
                    player, alsoAllowBlockDefs={endBlockDef})

                if cost is None:
                    # Never ending
                    continue

                endMapBlock = player.getMapBlock()
                if endMapBlock.defn not in (startBlockDef, endBlockDef):
                    # Left both blocks
                    continue

                if player.motionState == 'fall':
                    for actionClass in [FallDown, FallLeft, FallRight]:
                        options.append((steps + [actionClass], player.clone()))
                    continue

                if endMapBlock.defn == startBlockDef:
                    # Back in the starting block
                    continue

                finalNode = self.end.nodeFromPlayerState(player)
                newEdge = PreCalculatedEdge(
                    edge.startNode, steps, finalNode, edge.cost, orbTagCost)
                # Deliberately do not call newEdge.register() because the edge
                # belongs to this block combination, not just to the starting
                # block.
                edges.add(newEdge)
                directions = frozenset(finalNode.walkScores.iterkeys())
                partitions[directions] = partitions.get(directions, 0) + 1

        total = len(edges)
        self.edges = {}
        if edges:
            allDirections = frozenset(d for p in partitions for d in p)
            while allDirections not in partitions:
                worstPartition = min(partitions, key=partitions.get)
                log.warning(
                    '  WARNING: ignoring %r nodes exiting in %r',
                    partitions[worstPartition],
                    ''.join(sorted(worstPartition)))
                del partitions[worstPartition]
                allDirections = frozenset(d for p in partitions for d in p)

            bestPartition = allDirections

            for e in edges:
                if frozenset(e.endNode.walkScores.iterkeys()) == bestPartition:
                    oldEdge = self.edges.get(e.startNode)
                    if oldEdge is None or oldEdge.cost > e.cost:
                        self.edges[e.startNode] = e

            self.transitionCosts = {}
            for outDir in bestPartition:
                costs = [
                    e.endNode.walkScores[outDir]
                    for e in self.edges.itervalues()]
                self.transitionCosts[outDir] = (sum(costs) + 0.) / len(costs)

        log.info('  links: %r total, %r viable', total, len(edges))

    def calculateWalkScores(self):
        '''
        Calculates the walk scores for exiting the start block and ending up in
        the end block. Only stores values which differ from the default walk
        scores in the start block.
        '''

        heap = []
        for edge in self.edges.itervalues():
            heapq.heappush(heap, (edge.cost, edge.startNode))

        seen = set()
        while heap:
            cost, node = heapq.heappop(heap)
            if node in seen:
                continue
            seen.add(node)
            if cost != node.getWalkScore(self.direction):
                self.walkScores[node] = cost
            for edge in node.inbound:
                if edge.startNode in seen:
                    continue
                heapq.heappush(heap, (cost + edge.cost, edge.startNode))

        for node in self.start.nodes.itervalues():
            if node not in seen and self.direction in node.walkScores:
                self.walkScores[node] = None

        log.info(
            '  updated %r/%r nodes (/%r)', len(self.walkScores), len(seen),
            len(self.start.nodes))


class PreCalculatedNode(object):
    def __init__(self, blockPaths, block=None, obstacle=None, posIndex=None):
        self.blockPaths = blockPaths
        self.posIndex = posIndex

        if obstacle is None:
            self.obstacleId = None
            self.pos = None
        else:
            self.obstacleId = obstacle.obstacleId
            pos = obstacle.getPositionFromIndex(Player, posIndex)
            self.pos = (pos[0] - block.pos[0], pos[1] - block.pos[1])

        self.inbound = set()
        self.outbound = set()
        self.walkScores = {}

    def serialise(self):
        '''
        Returns an object that can be serialised using pickle, which contains
        all information about this node that is relevant at run-time.
        '''
        internalEdges = [
            e.serialise() for e in self.outbound
            if e.endNode.obstacleId is not None]

        targetEdges = {}
        orbEdges = [e for e in self.outbound if e.orbTagCost is not None]
        if orbEdges:
            bestOrbEdge = min(orbEdges, key=lambda e: e.orbTagCost)
            targetEdges[ORB] = bestOrbEdge.serialiseTargetEdge(ORB)

        return {
            'key': self.serialiseKey(),
            'edges': internalEdges,
            'targets': targetEdges,
            'scores': self.walkScores,
        }

    def serialiseKey(self):
        return (self.obstacleId, self.posIndex)

    def expandEdges(self, world, blockDef):
        assert not self.outbound

        obstacle = blockDef.getObstacleById(self.obstacleId)

        motionState = Player.getAttachedMotionState(obstacle)

        if motionState == 'fall':
            actionClasses = [FallDown, FallLeft, FallRight]
        elif motionState == 'leftwall':
            actionClasses = [Drop, JumpUp, JumpRight]
        elif motionState == 'rightwall':
            actionClasses = [Drop, JumpUp, JumpLeft]
        else:
            actionClasses = [
                JumpUp, JumpLeft, JumpRight,
                MoveLeft, MoveRight, SlowStepLeft, SlowStepRight]
            if obstacle.drop:
                actionClasses.append(Drop)

        for actionClass in actionClasses:
            self.expandAction(actionClass, world, blockDef, obstacle)

    def expandAction(self, actionClass, world, blockDef, obstacle):
        startPlayer = self.makePlayerState(world, blockDef, obstacle)
        startMapBlock = startPlayer.getMapBlock()

        options = [([actionClass], startPlayer)]
        while options:
            steps, player = options.pop(0)
            lastAction = steps[-1]
            cost, orbTagCost = lastAction.simulate(player)

            if cost is None:
                # Never ending
                continue

            endMapBlock = player.getMapBlock()
            if endMapBlock != startMapBlock:
                finalNode = self.blockPaths.getExitNode(
                    blockDef, endMapBlock.defn)
                if finalNode is None:
                    # Ended up in a non-adjacent map block
                    continue

            elif player.motionState != 'fall':
                finalNode = self.blockPaths.nodeFromPlayerState(player)
            else:
                for actionClass in [FallDown, FallLeft, FallRight]:
                    options.append((steps + [actionClass], player.clone()))
                continue

            edge = PreCalculatedEdge(self, steps, finalNode, cost, orbTagCost)
            edge.register()

    def makePlayerState(self, world, blockDef, obstacle=None):
        if obstacle is None:
            obstacle = blockDef.getObstacleById(self.obstacleId)

        player = Player(world, 'PathFinding', None, None)
        player.pos = (
            self.pos[0] + blockDef.pos[0], self.pos[1] + blockDef.pos[1])
        player.setAttachedObstacle(obstacle)
        return player

    def setWalkScore(self, scoreKey, cost):
        self.walkScores[scoreKey] = cost

    def getWalkScore(self, scoreKey):
        return self.walkScores[scoreKey]


class PreCalculatedEdge(object):
    def __init__(self, startNode, actions, endNode, cost, orbTagCost):
        self.startNode = startNode
        self.actions = actions
        self.endNode = endNode
        self.cost = cost
        self.orbTagCost = orbTagCost

    def register(self):
        self.startNode.outbound.add(self)
        self.endNode.inbound.add(self)
        if self.orbTagCost is not None:
            self.startNode.blockPaths.orbEdges.add(self)

    def serialise(self):
        return {
            'target': self.endNode.serialiseKey(),
            'cost': self.cost,
            'actions': [
                actionKindClasses.actionKindToString(a) for a in self.actions],
        }

    def serialiseTargetEdge(self, targetKey, includeSource=False):
        result = {
            'cost': self.orbTagCost if targetKey == ORB else self.cost,
            'actions': [
                actionKindClasses.actionKindToString(a) for a in self.actions],
        }
        if includeSource:
            result['source'] = self.startNode.serialiseKey()
        return result


pathFindingDBs = weakref.WeakKeyDictionary()


if __name__ == '__main__':
    from trosnoth.utils.utils import initLogging
    initLogging(debug=True)
    PathFindingStore().build()



import logging

from trosnoth.utils.message import UnhandledMessage
from trosnoth.utils.netmsg import NetworkMessage

log = logging.getLogger(__name__)


TICK_LIMIT = 1 << 16 - 1


####################
# Base Classes
####################

class AgentRequest(NetworkMessage):
    isControl = False
    timestampedPlayerRequest = False

    def clientValidate(self, localState, world, sendResponse):
        '''
        Called on the client to determine whether this message should even be
        passed on. Should return a boolean indication whether to continue to
        propagate this message.
        '''
        return True

    def applyRequestToLocalState(self, localState):
        '''
        Called on client->server messages before the message is sent to the
        server. This is used to update the local state information.
        '''
        pass

    def serverReceived(self, game, remoteAgent):
        '''
        Called server-side only for requests received over the network.
        '''
        remoteAgent.sendRequestToGame(self)

    def serverApply(self, game, agent):
        '''
        Called server-side for requests from agents.

        game:       a LocalGame
        agent:      an Agent

        Typically, a serverApply method will either do nothing, call
        game.sendServerCommand(), call game.sendResync(), or call
        agent.messageToAgent().
        '''
        raise UnhandledMessage('Validator did not expect to receive %s' % (
            self.__class__.__name__,))


class ServerCommand(NetworkMessage):
    isControl = False

    def applyOrderToWorld(self, world):
        '''
        Called when this order arrives at a Universe (on client and server) to
        update the state of the universe.
        '''
        pass

    def applyOrderToLocalState(self, localState, world):
        '''
        Called when this order arrives at the client, to update the local state
        of the client.
        '''
        pass


class ServerResponse(ServerCommand):
    isControl = True


class ClientCommand(AgentRequest, ServerCommand):
    '''
    A command that originates on the client and is validated on the server
    before being passed on.
    '''
    pass



from django.conf.urls import patterns, include, url
from django.contrib import admin

urlpatterns = patterns('',
    url(r'^', include('trosnoth.djangoapp.urls', namespace='trosnoth')),
    url(r'^admin/', include(admin.site.urls)),
)



import logging
import itertools
import string
import types

# timeNow is used to update things based on how much time has passed.
# Note: on Windows, time.clock() is more precise than time.time()
# On Windows time.clock() also does not change when the system clock changes.
# On linux however, time.clock() measures process time rather than wall time.
import platform
if platform.system() == 'Windows':
    from time import clock as timeFunction
else:
    from time import time as timeFunction

from hashlib import sha1 as hasher


def timeNow():
    '''
    This function exists so that even things which get a reference to timeNow()
    at import time can still be fooled by patching timeFunction.
    '''
    return timeFunction()


def new(count):
    '''new(count) - returns an iterator object which will give count distinct
    instances of the object class.  This is useful for defining setting
    options.  For example, north, south, east, west = new(4) . There is no
    reason that these options should be given numeric values, but it is
    important that north != south != east != west.
    '''
    for i in xrange(count):
        yield object()


class RunningAverage(object):
    def __init__(self, count):
        self.values = []
        self.total = 0.
        self.maxCount = count

    @property
    def mean(self):
        if len(self.values) == 0:
            return None
        return self.total / len(self.values)

    def noteValue(self, value):
        self.total += value
        self.values.append(value)
        if len(self.values) > self.maxCount:
            self.total -= self.values.pop(0)


class StaticMeta(type):
    '''
    Metaclass for Function and Static classes below. Ensures that all methods
    are class methods.
    '''
    def __new__(cls, name, bases, dict_):
        for k, v in dict_.iteritems():
            if k == '__new__':
                continue
            if isinstance(v, types.FunctionType):
                dict_[k] = classmethod(v)
        return super(StaticMeta, cls).__new__(cls, name, bases, dict_)


class Static(object):
    '''
    Superclass for defining classes that are not meant to be instantiated. All
    methods become class methods by default.
    '''
    __metaclass__ = StaticMeta

    def __new__(cls, *args, **kwargs):
        raise NotImplementedError(
            '%s class cannot be instantiated' % (cls.__name__,))


def initLogging(debug=False, logFile=None):
    import twisted.python.log
    if debug:
        logging.getLogger().setLevel(logging.DEBUG)
    h = logging.StreamHandler()
    h.setFormatter(logging.Formatter('%(relativeCreated)7d %(message)s'))
    logging.getLogger().addHandler(h)
    if logFile:
        h = logging.FileHandler(logFile)
        h.setFormatter(logging.Formatter('%(asctime)s %(message)s'))
    observer = twisted.python.log.PythonLoggingObserver()
    twisted.python.log.startLoggingWithObserver(observer.emit)


# Convenience functions for wrapping long strings based on maximum pixel width
# http://www.pygame.org/wiki/TextWrapping

def truncline(text, font, maxwidth):
    real = len(text)
    stext = text
    l = font.size(text)[0]
    cut = 0
    a = 0
    done = 1
    while l > maxwidth:
        a = a + 1
        n = text.rsplit(None, a)[0]
        if stext == n:
            cut += 1
            stext = n[:-cut]
        else:
            stext = n
        l = font.size(stext)[0]
        real = len(stext)
        done = 0
    return real, done, stext


def wrapline(text, font, maxwidth):
    done = 0
    wrapped = []

    while not done:
        nl, done, stext = truncline(text, font, maxwidth)
        wrapped.append(stext.strip())
        text = text[nl:]
    return wrapped


def wrapMultiLine(text, font, maxwidth):
    lines = itertools.chain(*(
        wrapline(line, font, maxwidth) for line in text.splitlines()))
    return list(lines)


class BasicContextManager(object):
    def __enter__(self):
        return self

    def __exit__(self, type, value, traceback):
        pass


def stripPunctuation(nick):
    exclude = set(string.punctuation + ' ')
    return ''.join(ch for ch in nick if ch not in exclude)



import logging

from trosnoth.const import (
    GAME_FULL_REASON, UNAUTHORISED_REASON, NICK_USED_REASON, BAD_NICK_REASON,
    USER_IN_GAME_REASON, ALREADY_JOINED_REASON,
)
from trosnoth.messages.base import (
    AgentRequest, ServerCommand, ServerResponse, ClientCommand,
)
from trosnoth.messages.special import PlayerHasElephantMsg
from trosnoth.model.universe_base import NeutralTeamId
from trosnoth.utils.netmsg import NetworkMessage

log = logging.getLogger(__name__)


####################
# Setup
####################

class InitClientMsg(NetworkMessage):
    idString = 'wlcm'
    fields = 'settings'
    packspec = '*'


class ConnectionLostMsg(ServerCommand):
    # Doesn't really originate on the server.
    fields = ()
    isControl = True
    packspec = ''

    def applyOrderToWorld(self, world):
        pass


class WorldResetMsg(ServerCommand):
    idString = 'rset'
    fields = 'settings'
    packspec = '*'


class ZoneStateMsg(ServerCommand):
    idString = 'ZnSt'
    fields = 'zoneId', 'teamId', 'dark'
    packspec = 'Icb'


class WorldLoadingMsg(ServerCommand):
    idString = 'wait'
    fields = 'loading'
    packspec = 'b'

    def applyOrderToWorld(self, world):
        world.loading = self.loading


####################
# Game
####################

class ChangeNicknameMsg(ClientCommand):
    idString = 'Nick'
    fields = 'playerId', 'nickname'
    packspec = 'c*'

    def serverApply(self, game, agent):
        if game.world.gameIsInProgress():
            return
        p = game.world.getPlayer(self.playerId)
        if not p:
            return

        nick = self.nickname.decode()
        if not game.world.isValidNick(nick):
            return

        for player in game.world.players:
            if p != player and player.nick.lower() == nick.lower():
                # Nick in use.
                return

        if game.world.authManager is not None:
            if game.world.authManager.checkUsername(nick) and (
                    p.user is None or nick.lower() != p.user.username):
                # Nick is someone else's username.
                return

        game.sendServerCommand(self)
        if p.user:
            p.user.setNick(nick)

    def applyOrderToWorld(self, world):
        player = world.getPlayer(self.playerId)
        if player:
            player.nick = self.nickname.decode()

    def applyOrderToLocalState(self, localState, world):
        player = localState.player
        if player and player.id == self.playerId:
            player.nick = self.nickname.decode()


class PlayerIsReadyMsg(ClientCommand):
    idString = 'Redy'
    fields = 'playerId', 'ready', 'tournament'
    packspec = 'c??'

    def serverApply(self, game, agent):
        game.sendServerCommand(self)

    def applyOrderToWorld(self, world):
        self.applyToPlayer(world.getPlayer(self.playerId))

    def applyOrderToLocalState(self, localState, world):
        player = localState.player
        if player and player.id == self.playerId:
            self.applyToPlayer(player)

    def applyToPlayer(self, player):
        player.readyToStart = self.ready
        player.readyForTournament = (
            self.tournament and player.prefersTournamentTeam)


class SetPreferredTeamMsg(AgentRequest):
    idString = 'TmNm'
    fields = 'name'
    packspec = '*'

    MAX_TEAM_NAME_LENGTH = 30

    def serverApply(self, game, agent):
        if agent.player is None:
            return
        teamName = self.name.decode()[:self.MAX_TEAM_NAME_LENGTH]
        game.sendServerCommand(PreferredTeamSelectedMsg(
            agent.player.id,
            game.world.isTournamentTeam(teamName),
            teamName.encode()))


class PreferredTeamSelectedMsg(ServerCommand):
    idString = 'TmNm'
    fields = 'playerId', 'tournament', 'name'
    packspec = 'c?*'

    def applyOrderToWorld(self, world):
        self.applyToPlayer(world.getPlayer(self.playerId))

    def applyOrderToLocalState(self, localState, world):
        player = localState.player
        if player and player.id == self.playerId:
            self.applyToPlayer(player)

    def applyToPlayer(self, player):
        player.preferredTeam = self.name.decode()
        player.prefersTournamentTeam = self.tournament
        if not player.prefersTournamentTeam:
            player.readyForTournament = False


class SetPreferredSizeMsg(ClientCommand):
    idString = 'Size'
    fields = 'playerId', 'halfMapWidth', 'mapHeight'
    packspec = 'cBB'

    MAX_MAP_HEIGHT = 10
    MAX_HALF_WIDTH = 10

    def serverApply(self, game, agent):
        if game.world.getPlayer(self.playerId):
            halfMapWidth = min(self.halfMapWidth, self.MAX_HALF_WIDTH)
            mapHeight = min(self.mapHeight, self.MAX_MAP_HEIGHT)
            game.sendServerCommand(
                SetPreferredSizeMsg(self.playerId, halfMapWidth, mapHeight))

    def applyOrderToWorld(self, world):
        self.applyToPlayer(world.getPlayer(self.playerId))

    def applyOrderToLocalState(self, localState, world):
        player = localState.player
        if player and player.id == self.playerId:
            self.applyToPlayer(player)

    def applyToPlayer(self, player):
        player.preferredSize = (self.halfMapWidth, self.mapHeight)


class SetPreferredDurationMsg(ClientCommand):
    '''
    duration is in seconds.
    '''
    idString = 'Dr8n'
    fields = 'playerId', 'duration'
    packspec = 'cI'

    MAX_GAME_DURATION = 86400

    def serverApply(self, game, agent):
        if game.world.getPlayer(self.playerId):
            duration = min(self.duration, self.MAX_GAME_DURATION)
            game.sendServerCommand(
                SetPreferredDurationMsg(self.playerId, duration))

    def applyOrderToWorld(self, world):
        self.applyToPlayer(world.getPlayer(self.playerId))

    def applyOrderToLocalState(self, localState, world):
        player = localState.player
        if player and player.id == self.playerId:
            self.applyToPlayer(player)

    def applyToPlayer(self, player):
        player.preferredDuration = self.duration


class GameStartMsg(ServerCommand):
    idString = 'Go!!'
    fields = ()
    packspec = ''


class GameOverMsg(ServerCommand):
    idString = 'Stop'
    fields = 'teamId', 'timeOver'
    packspec = 'cB'


class ReturnToLobbyMsg(ServerCommand):
    idString = 'Loby'
    fields = ()
    packspec = ''


class SetGameModeMsg(ServerCommand):
    idString = 'Mode'
    fields = 'gameMode'
    packspec = '*'

    def applyOrderToWorld(self, world):
        world.setGameMode(self.gameMode.decode())


class SetGameSpeedMsg(ServerCommand):
    idString = 'Spee'
    fields = 'gameSpeed'
    packspec = 'f'

    def applyOrderToWorld(self, world):
        world.setGameSpeed(self.gameSpeed)


class SetTeamNameMsg(ServerCommand):
    idString = 'Team'
    fields = 'teamId', 'name'
    packspec = 'c*'

    def applyOrderToWorld(self, world):
        if self.teamId == '\x00':
            world.rogueTeamName = self.name
        else:
            team = world.getTeam(self.teamId)
            team.teamName = self.name


class StartingSoonMsg(ServerCommand):
    idString = 'Wait'
    fields = ('delay', 'gameType')
    packspec = 'd*'


class ChangeTimeLimitMsg(ServerCommand):
    idString = 'Time'
    fields = 'timeLimit'
    packspec = 'd'


####################
# Players
####################

class AddPlayerMsg(ServerCommand):
    idString = 'NewP'
    fields = 'playerId', 'teamId', 'zoneId', 'dead', 'bot', 'nick'
    packspec = 'ccIbb*'


class SetPlayerTeamMsg(ServerCommand):
    idString = 'PlTm'
    fields = 'playerId', 'teamId'
    packspec = 'cc'

    def applyOrderToWorld(self, world):
        player = world.getPlayer(self.playerId)
        player.team = world.getTeam(self.teamId)

    def applyOrderToLocalState(self, localState, world):
        if localState.player and localState.player.id == self.playerId:
            localState.player.team = world.getTeam(self.teamId)


class RemovePlayerMsg(ServerCommand):
    idString = 'DelP'
    fields = 'playerId'
    packspec = 'c'

    def applyOrderToWorld(self, world):
        world.delPlayer(world.getPlayer(self.playerId))


class JoinRequestMsg(AgentRequest):
    idString = 'Join'
    fields = 'teamId', 'authTag', 'bot', 'nick'
    packspec = 'cQb*'
    authTag = 0
    localBotRequest = False     # Intentionally not sent over wire.

    def serverApply(self, game, agent):
        preferredTeamId = self.teamId
        nick = self.nick.decode()
        teamPlayerCounts = game.world.getTeamPlayerCounts()

        if agent.player is not None:
            agent.messageToAgent(CannotJoinMsg(ALREADY_JOINED_REASON))
            return

        if len(game.world.players) >= game.maxTotalPlayers:
            agent.messageToAgent(CannotJoinMsg(GAME_FULL_REASON))
            return

        teamId = game.world.getTeamToJoin(preferredTeamId)
        if (teamId != NeutralTeamId and teamPlayerCounts.get(teamId, 0) >=
                game.maxPerTeam):
            agent.messageToAgent(CannotJoinMsg(GAME_FULL_REASON))
            return

        if not game.world.isValidNick(nick):
            agent.messageToAgent(CannotJoinMsg(BAD_NICK_REASON))
            return

        preferredTeam = ''
        if game.authTagManager is not None and not self.localBotRequest:
            user = game.authTagManager.checkAuthTag(self.authTag)
            if user is None:
                agent.messageToAgent(CannotJoinMsg(UNAUTHORISED_REASON))
                return
            else:
                for player in game.world.players:
                    if player.user == user:
                        agent.messageToAgent(
                            CannotJoinMsg(USER_IN_GAME_REASON))
                        return
                if (nick.lower() != user.username and
                        game.authTagManager.checkUsername(nick)):
                    agent.messageToAgent(CannotJoinMsg(NICK_USED_REASON))
                    return

                preferredTeam = user.getDefaultTeam()
        else:
            user = None

        # Only check for duplicate nick after checking for auth-related errors.
        usedNicks = {player.nick.lower() for player in game.world.players}
        if self.bot:
            if '-' in nick and nick.rsplit('-', 1)[-1].isdigit():
                base, number = nick.rsplit('-', 1)
                number = int(number)
            else:
                base = nick
                number = 0

            if 'Bot' not in base:
                base += 'Bot'

            while nick.lower() in usedNicks:
                number += 1
                nick = '{}-{}'.format(base, number)
        else:
            if nick.lower() in usedNicks:
                agent.messageToAgent(CannotJoinMsg(NICK_USED_REASON))
                return

        if game.authTagManager is not None and not self.localBotRequest:
            game.authTagManager.discardAuthTag(self.authTag)
            user.setNick(nick)

        playerId = game.idManager.newPlayerId()
        if playerId is None:
            agent.messageToAgent(CannotJoinMsg(GAME_FULL_REASON))
            return

        zoneId = game.world.selectZoneForTeam(teamId).id
        game.sendServerCommand(AddPlayerMsg(
            playerId, teamId, zoneId, True, self.bot, nick.encode()))
        player = game.world.getPlayer(playerId)
        if user:
            player.setUser(user)
        if player.isElephantOwner():
            game.world.sendServerCommand(PlayerHasElephantMsg(playerId))
        game.joinSuccessful(agent, playerId)

        if preferredTeam:
            tournament = game.world.isTournamentTeam(preferredTeam)
            game.sendServerCommand(PreferredTeamSelectedMsg(
                playerId, tournament, preferredTeam.encode()))


class CannotJoinMsg(ServerResponse):
    '''
    Valid reasonId options are defined in trosnoth.const.
    '''
    idString = 'NotP'
    fields = 'reasonId'
    packspec = 'c'


class SetAgentPlayerMsg(ServerCommand):
    '''
    Send back to the agent which requested to join if the join has succeeded.

    This is not a control message, because we only want the successful join to
    be delivered after the player has been added to the universe.
    '''
    idString = 'OwnP'
    fields = 'playerId'
    packspec = 'c'

    def applyOrderToLocalState(self, localState, world):
        localState.agent.setPlayer(world.getPlayer(self.playerId))



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

import datetime
import json
import os

from django.db import migrations

from trosnoth import data




def migrate_old_user_info(apps, schema_editor):
    User = apps.get_registered_model('auth', 'User')
    TrosnothUser = apps.get_model('trosnoth', 'TrosnothUser')
    AchievementProgress = apps.get_model('trosnoth', 'AchievementProgress')

    dataPath = data.getPath(data.user, 'authserver')
    path = os.path.join(dataPath, 'accounts')
    for username in os.listdir(path):
        userpath = os.path.join(path, username)
        user = User(username=username, email='')
        user.save()

        with open(os.path.join(userpath, 'nick'), 'rb') as f:
            nick = f.read().decode()
        with open(os.path.join(userpath, 'stats'), 'rU') as f:
            stats = json.load(f)
        with open(os.path.join(userpath, 'password'), 'rb') as f:
            passwordHash = f.read()
        lastSeen = datetime.datetime.fromtimestamp(stats.get('lastSeen'))
        trosnothUser = TrosnothUser(
            user=user, nick=nick, lastSeen=lastSeen,
            oldPasswordHash=passwordHash)
        trosnothUser.save()

        achievementPath = os.path.join(userpath, 'achievements')
        if os.path.exists(achievementPath):
            with open(os.path.join(userpath, 'achievements'), 'rU') as f:
                achievements = json.load(f)

            for achId, achData in achievements.iteritems():
                progress = None
                extra = ''
                if 'progress' in achData:
                    if isinstance(achData['progress'], int):
                        progress = achData['progress']
                    else:
                        extra = json.dumps(achData['progress'])

                record = AchievementProgress(
                    user=trosnothUser, achievementId=achId,
                    unlocked=achData.get('unlocked', False),
                    progress=progress, data=extra)
                record.save()


class Migration(migrations.Migration):

    dependencies = [
        ('trosnoth', '0001_initial'),
        ('auth', '0001_initial'),
    ]

    operations = [
        migrations.RunPython(migrate_old_user_info),
    ]



'''miniMap.py - defines the MiniMap class which deals with drawing the
miniMap to the screen.'''

import random
import logging
import pygame
from twisted.internet import reactor

from trosnoth.const import MAP_TO_SCREEN_SCALE
from trosnoth.model.map import MapLayout
from trosnoth.model import mapblocks
import trosnoth.gui.framework.framework as framework
from trosnoth.gui.framework.basics import Animation
from trosnoth.trosnothgui.ingame.sprites import PlayerSprite
from trosnoth.utils.utils import timeNow

log = logging.getLogger('minimap')

SHADOW_OFFSET = 1, 1


class MiniMap(framework.Element):

    def __init__(self, app, scale, universeGui, viewManager):
        '''Called upon creation of a ViewManager object.  screen is a pygame
        screen object.  universe is the Universe object to draw.  target is
        either a point, a PlayerSprite object, or None.  If target is None, the
        view manager will follow the action, otherwise it will follow the
        specified point or player.'''
        super(MiniMap, self).__init__(app)
        self.universeGui = universeGui
        self.universe = universeGui.universe
        self.viewManager = viewManager
        self.disrupt = False
        # Initialise the graphics in all the mapBlocks
        bodyBlockSize = (
            int(MapLayout.zoneBodyWidth / scale + 0.5),
            int(MapLayout.halfZoneHeight / scale + 0.5))
        self.bodyBlockRect = pygame.Rect((0,0), bodyBlockSize)

        interfaceBlockSize = (
            int(MapLayout.zoneInterfaceWidth / scale + 0.5),
            int(MapLayout.halfZoneHeight / scale + 0.5))
        self.interfaceBlockRect = pygame.Rect((0,0), interfaceBlockSize)

        # Recalculate scale based on integer-sized block rectangles
        self.scale = (
            (MapLayout.zoneBodyWidth + MapLayout.zoneInterfaceWidth + 0.)
            / (bodyBlockSize[0] + interfaceBlockSize[0]))
        self.graphicsScale = self.scale * MAP_TO_SCREEN_SCALE

        self.disruptAnimation = self.createDisrupted()

        # self._focus represents the point where the miniMap is currently
        # looking.
        self._focus = None
        self.updateFocus()

    def getScaledMaximumSize(self):
        # 35% of screen size
        return tuple([self.app.screenManager.scaledSize[i] * 35 / 100
                for i in (0, 1)])

    def getAbsoluteMaximumSize(self):
        # Could be aptly described as the minimum maximum
        return (300,200)

    def getMaximumSize(self):
        size1 = self.getScaledMaximumSize()
        size2 = self.getAbsoluteMaximumSize()
        return tuple([max(size1[i], size2[i]) for i in 0,1])

    def getUniverseScaledSize(self):
        universeSize = (len(self.universe.zoneBlocks[0]),
                len(self.universe.zoneBlocks))

        mapHeight = universeSize[1] * self.bodyBlockRect.height
        mapWidth = ((universeSize[0] / 2) * self.bodyBlockRect.height +
                (universeSize[0] / 2 + 1) * self.interfaceBlockRect.width)

        return (mapWidth, mapHeight)

    def getSize(self):
        size1 = self.getUniverseScaledSize()
        size2 = self.getMaximumSize()
        return tuple([min(size1[i], size2[i]) for i in 0,1])

    def getOffset(self):
        return self.app.screenManager.size[0] - self.getSize()[0] - 5, 5

    def getRect(self):
        return pygame.Rect(self.getOffset(), self.getSize())

    def createDisrupted(self):
        sRect = self.getRect()
        screens = [pygame.surface.Surface(sRect.size) for i in xrange(4)]

        # Because the creation of these disruption sprites is very slow, do
        # them a bit at a time so that other events don't get blocked.
        for screen in screens:
            reactor.callLater(0, self.drawSingleDisruption, screen)
        return Animation(0.1, timeNow, *screens)

    def drawSingleDisruption(self, screen, y=0, sRect=None, rect=None):
        if sRect is None:
            sRect = self.getRect()
        if rect is None:
            rect = pygame.rect.Rect(0, 0, 2, 2)
        disruptColours = [
            self.app.theme.colours.minimapDisruptionColour1,
            self.app.theme.colours.minimapDisruptionColour2
        ]

        rows = 0
        while y < sRect.bottom:
            x = 0
            rect.top = y
            while x < sRect.right:
                rect.left = x
                pygame.draw.rect(screen, random.choice(disruptColours),
                        rect, 0)
                x += 2
            y += 2
            rows += 1
            if rows >= 5:
                reactor.callLater(
                    0, self.drawSingleDisruption, screen, y, sRect, rect)
                break

    def drawDisrupted(self, screen):
        screen.blit(self.disruptAnimation.getImage(), self.getOffset())

    def mapPosToMinimap(self, sRect, (x, y)):
        fx, fy = self._focus
        cx, cy = sRect.center
        s = self.scale
        return (int((x - fx + 0.5) / s + cx), int((y - fy + 0.5) / s + cy))

    def minimapPosToMap(self, sRect, (x, y)):
        fx, fy = self._focus
        cx, cy = sRect.center
        s = self.scale
        return ((x - cx) * s + fx, (y - cy) * s + fy)

    def draw(self, screen):
        '''Draws the current state of the universe at the current viewing
        location on the screen.  Does not call pygame.display.flip() .'''
        sRect = self.getRect()

        oldClip = screen.get_clip()
        screen.set_clip(sRect)

        colours = self.app.theme.colours
        pygame.draw.rect(screen, colours.black, sRect, 0)

        # If disrupted, draw static 95% of the time
        if self.disrupt and random.random() > 0.05:
            self.drawDisrupted(screen)
        else:
            # Update where we're looking at.
            self.updateFocus()
            if self.universe.uiOptions.shiftingFrontLine:
                self.drawShiftingBg(screen, sRect)
            else:
                self.drawZones(screen, sRect)

            def drawCircle(sprite, colour, radius):
                pos = self.mapPosToMinimap(sRect, sprite.pos)
                if sRect.collidepoint(*pos):
                    pygame.draw.circle(screen, colour, pos, radius)

            # Draw the shots
            for shotSprite in self.universeGui.iterShots():
                shot = shotSprite.shot
                if not shot.expired:
                    drawCircle(shotSprite, colours.white, 0)
            # Draw the stars
            for star in self.universe.collectableStars.values():
                drawCircle(star, colours.miniMapStar, 2)

            # Draw the trosball
            sprite = self.universeGui.getTrosballSprite()
            if sprite:
                drawCircle(sprite, colours.white, 4)

            # Go through and update the positions of the players on the screen.
            for playerSprite in self.universeGui.iterPlayers():
                player = playerSprite.player
                # The player being drawn is the one controlled by the user.
                if playerSprite == self.viewManager.target:
                    clr = colours.minimapOwnColour
                    radius = 3

                else:
                    if player.dead:
                        clr = colours.miniMapGhostColour(player.team)
                    else:
                        clr = colours.miniMapPlayerColour(player.team)
                    radius = 2

                drawCircle(playerSprite, clr, radius)

            if self.universe.uiOptions.showNets:
                for team in self.universe.teams:
                    colour = colours.miniMapTrosballTargetColour(team)
                    drawCircle(self.universe.getTrosballTargetZoneDefn(team),
                            colour, 3)

        # Finally, draw the border
        screen.set_clip(oldClip)
        pygame.draw.rect(screen, colours.minimapBorder, sRect, 2)

    def screenToMap(self, pt):
        '''
        Converts the given point to a map position assuming it's inside the
        minimap's area.
        '''
        sRect = self.getRect()
        topCorner = [self._focus[i] - (sRect.size[i] / 2 * self.scale)
                     for i in (0,1)]
        mapOffset = ((pt[0] - sRect.left) * self.scale, (pt[1] - sRect.top) *
                self.scale)
        return (topCorner[0] + mapOffset[0], topCorner[1] + mapOffset[1])

    def drawShiftingBg(self, screen, sRect):
        try:
            frontLine = self.universe.uiOptions.frontLineGetter()
        except ValueError:
            self.drawZones(screen, sRect)
            return

        minimapTrosballPosition = self.mapPosToMinimap(
            sRect, (frontLine, 0))[0]
        colours = self.app.theme.colours
        blue = colours.miniMapDarkZoneColour(self.universe.teams[0])
        red = colours.miniMapDarkZoneColour(self.universe.teams[1])
        if minimapTrosballPosition > sRect.right:
            self.drawZones(screen, sRect, forceColour=blue)
        elif minimapTrosballPosition < sRect.left:
            self.drawZones(screen, sRect, forceColour=red)
        else:
            blueWidth = minimapTrosballPosition - sRect.left
            blueRect = pygame.Rect(sRect.left, sRect.top, blueWidth,
                    sRect.height)
            redRect = pygame.Rect(minimapTrosballPosition, sRect.top,
                    sRect.width - blueWidth, sRect.height)
            clip = screen.get_clip()
            try:
                screen.set_clip(blueRect)
                self.drawZones(screen, sRect, forceColour=blue)
                screen.set_clip(redRect)
                self.drawZones(screen, sRect, forceColour=red)
            finally:
                screen.set_clip(clip)

    def drawZones(self, screen, sRect, forceColour=None):
        '''Draws the miniMap graphics onto the screen'''
        topCorner = [self._focus[i] - (sRect.size[i] / 2 * self.scale)
                     for i in (0,1)]
        # Find which map blocks are on the screen.
        i, j = MapLayout.getMapBlockIndices(*topCorner)
        i = max(0, i)
        j = max(0, j)
        firstBlock = self.universe.zoneBlocks[i][j]


        # Set the initial position back to where it should be.
        firstPos = self.mapPosToMinimap(sRect, firstBlock.defn.rect.topleft)
        firstPos = [min(firstPos[a], sRect.topleft[a]) for a in (0, 1)]

        posToDraw = [firstPos[a] for a in (0,1)]
        y, x = i, j
        while posToDraw[1] < sRect.bottom:
            while posToDraw[0] < sRect.right:
                try:
                    block = self.universe.zoneBlocks[y][x]
                except IndexError:
                    break
                zone = None
                if isinstance(block, mapblocks.InterfaceMapBlock):
                    currentRect = self.interfaceBlockRect
                elif isinstance(block, mapblocks.BottomBodyMapBlock):
                    currentRect = self.bodyBlockRect
                    zone = block.zone
                else:
                    currentRect = self.bodyBlockRect
                currentRect.topleft = posToDraw
                area = currentRect.clip(sRect)
                draw = True
                if area.size == currentRect.size:
                    # Nothing has changed.
                    self._drawBlockMiniBg(
                        block, screen, sRect, currentRect,
                        forceColour=forceColour)
                elif area.size == (0,0):
                    # Outside the bounds of the minimap
                    draw = False
                else:
                    self._drawBlockMiniBg(
                        block, screen, sRect, currentRect, area,
                        forceColour=forceColour)

                if draw and zone:
                    self.drawZoneDecoration(
                        zone, screen, sRect, currentRect.midtop)

                posToDraw[0] += currentRect.width
                x += 1
            x = j
            y += 1
            # Next Row
            posToDraw[0] = firstPos[0]
            posToDraw[1] += self.interfaceBlockRect.height

    def drawZoneLetter(self, screen, sRect, zone, centre):
        img = self.renderLetter(zone.defn.label)
        rect = img.get_rect()
        rect.center = centre
        screen.blit(img, rect)

    def getZoneHighlight(self, zone):
        '''
        Determines whether the given zone should be highlighted and if so in
        what colour. Returns None if the zone should not be highlighted on the
        minimap, otherwise returns the highlight colour.
        '''
        sprites = self.app.theme.sprites

        player = self.getTargetPlayer()
        if player is None or player.team is None:
            playerTeam = self.universe.teams[0]
        else:
            playerTeam = player.team
        owner = zone.owner
        if zone.isCapturableBy(playerTeam):
            return sprites.zoneHighlight(playerTeam, self.scale)
        if owner == playerTeam or owner is None:
            teams = zone.teamsAbleToTag()
            teams.discard(playerTeam)
            if teams:
                return sprites.zoneHighlight(teams.pop(), self.scale)
        if zone.isDisputed():
            return sprites.zoneHighlight(None, self.scale)
        return None

    def drawZoneDecoration(self, zone, screen, sRect, centre):
        if not self.universe.uiOptions.showNets:
            highlight = self.getZoneHighlight(zone)
            if highlight:
                rect = highlight.get_rect()
                rect.center = centre
                screen.blit(highlight, rect)
        self.drawZoneLetter(screen, sRect, zone, centre)

    def _drawBlockMiniBg(self, block, surface, sRect, rect, area=None,
            forceColour=None):
        if block.defn.kind == 'fwd':
            self._drawFwdBlockMiniBg(block, surface, rect, area, forceColour)
        elif block.defn.kind == 'bck':
            self._drawBckBlockMiniBg(block, surface, rect, area, forceColour)
        else:
            self._drawBodyBlockMiniBg(block, surface, rect, area, forceColour)
        self._drawBlockMiniArtwork(block, surface, rect, area)

    def _drawBlockMiniArtwork(self, block, surface, rect, area):
        if block.defn.graphics is None:
            return

        if area is not None:
            cropPos = (area.left - rect.left, area.top - rect.top)
            crop = pygame.rect.Rect(cropPos, area.size)
            surface.blit(block.defn.graphics.getMini(
                self.app, self.graphicsScale), area.topleft, crop)
        else:
            surface.blit(block.defn.graphics.getMini(
                self.app, self.graphicsScale), rect.topleft)

    def _drawBodyBlockMiniBg(self, block, surface, rect, area,
            forceColour=None):
        if block.zone:
            clr = (self._getMiniMapColour(block.zone) if forceColour is None
                    else forceColour)
        else:
            clr = (0, 0, 0)
        if area is not None:
            pygame.draw.rect(surface, clr, area)
        else:
            pygame.draw.rect(surface, clr, rect)

    def renderLetter(self, letter):
        if not hasattr(self, 'letters'):
            self.letters = {}
        if letter not in self.letters:
            font = self.app.screenManager.fonts.ingameMenuFont
            shadow = font.render(self.app, letter, False, (0, 0, 0))
            highlight = font.render(self.app, letter, False, (255, 255, 255))
            x, y = highlight.get_size()
            xOff, yOff = SHADOW_OFFSET
            result = pygame.Surface((x + xOff, y + yOff)).convert()
            result.fill((0, 0, 1))
            result.set_colorkey((0, 0, 1))
            result.blit(shadow, SHADOW_OFFSET)
            result.blit(highlight, (0, 0))
            self.letters[letter] = result
        return self.letters[letter]

    def _drawFwdBlockMiniBg(self, block, surface, rect, area, forceColour=None):
        if area:
            tempSurface = pygame.surface.Surface(rect.size)
            tempRect = tempSurface.get_rect()
        if block.zone1:
            clr = (self._getMiniMapColour(block.zone1) if forceColour is None
                    else forceColour)
        else:
            clr = (0, 0, 0)
        if area:
            pts = (tempRect.topleft, tempRect.topright, tempRect.bottomleft)
            pygame.draw.polygon(tempSurface, clr, pts)
        else:
            pts = (rect.topleft, rect.topright, rect.bottomleft)
            pygame.draw.polygon(surface, clr, pts)

        if block.zone2:
            clr = (self._getMiniMapColour(block.zone2) if forceColour is None
                    else forceColour)
        else:
            clr = (0, 0, 0)

        if area:
            pts = (tempRect.bottomright, tempRect.topright, tempRect.bottomleft)
            pygame.draw.polygon(tempSurface, clr, pts)
            # Now put it onto surface
            cropPos = (area.left - rect.left, area.top - rect.top)
            crop = pygame.rect.Rect(cropPos, area.size)
            surface.blit(tempSurface, area.topleft, crop)
        else:
            pts = (rect.bottomright, rect.topright, rect.bottomleft)
            pygame.draw.polygon(surface, clr, pts)

    def _drawBckBlockMiniBg(self, block, surface, rect, area, forceColour=None):
        if area:
            tempSurface = pygame.surface.Surface(rect.size)
            tempRect = tempSurface.get_rect()
        if block.zone1:
            clr = (self._getMiniMapColour(block.zone1) if forceColour is None
                    else forceColour)
        else:
            clr = (0, 0, 0)

        if area:
            pts = (tempRect.topleft, tempRect.bottomright, tempRect.bottomleft)
            pygame.draw.polygon(tempSurface, clr, pts)
        else:
            pts = (rect.topleft, rect.bottomright, rect.bottomleft)
            pygame.draw.polygon(surface, clr, pts)

        if block.zone2:
            clr = (self._getMiniMapColour(block.zone2) if forceColour is None
                    else forceColour)
        else:
            clr = (0, 0, 0)
        if area:
            pts = (tempRect.topleft, tempRect.bottomright, tempRect.topright)
            pygame.draw.polygon(tempSurface, clr, pts)
            # Now put it onto surface
            cropPos = (area.left - rect.left, area.top - rect.top)
            crop = pygame.rect.Rect(cropPos, area.size)
            surface.blit(tempSurface, area.topleft, crop)
        else:
            pts = (rect.topleft, rect.bottomright, rect.topright)
            pygame.draw.polygon(surface, clr, pts)

    def _getMiniMapColour(self, zone):
        colours = self.app.theme.colours
        if zone.dark:
            result = colours.miniMapDarkZoneColour(zone.owner)
        else:
            result = colours.miniMapLightZoneColour(zone.owner)
        return result

    # The right-most and left-most positions at which the minimap can focus
    def getBoundaries(self):
            # The edge of the map will always be an interfaceMapBlock
        indices = (len(self.universe.zoneBlocks) - 1,
                   len(self.universe.zoneBlocks[0]) - 1)

        block = self.universe.zoneBlocks[indices[0]][indices[1]]
        pos = block.defn.rect.bottomright
        sRect = self.getRect()
        rightMost = (pos[0] - sRect.size[0] * self.scale / 2,
                             pos[1] - sRect.size[1] * self.scale / 2)

        # The left-most position at which the minimap can focus
        leftMost = (sRect.size[0] * self.scale / 2,
                            sRect.size[1] * self.scale / 2)

        return rightMost, leftMost

    def getTargetPlayer(self):
        if isinstance(self.viewManager.target, PlayerSprite):
            return self.viewManager.target.player
        return None

    def updateFocus(self):
        if isinstance(self.viewManager.target, PlayerSprite):
            self._focus = self.viewManager.target.pos
        elif self.viewManager.target is None:
            self._focus = self.viewManager._focus
        else:
            #assert isinstance(self._focus, tuple)
            self._focus = self.viewManager.target

        rightMost, leftMost = self.getBoundaries()
        self._focus = [max(min(self._focus[i], rightMost[i]), leftMost[i]) for
                i in (0,1)]

    def disrupted(self):
        self.disrupt = True

    def endDisruption(self):
        self.disrupt = False

    def getZoneAtPoint(self, pt):
        if self.getRect().collidepoint(pt):
            x, y = self.screenToMap(pt)
            i, j = MapLayout.getMapBlockIndices(x, y)
            try:
                return self.universe.map.zoneBlocks[i][j].getZoneAtPoint(x, y)
            except IndexError:
                return None
        return None



# Trosnoth (UberTweak Platform Game)
# Copyright (C) 2006-2012 Joshua D Bartlett
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# version 2 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301, USA.

from math import pi
import pygame


class ImageCollection(object):
    def __init__(self, *images):
        for image in images:
            if not isinstance(image, pygame.Surface):
                raise TypeError('%r is not a pygame.Surface' % (image,))
        self.images = images

    def getImage(self):
        raise NotImplementedError('getImage not implemented')


class SingleImage(ImageCollection):
    def __init__(self, image):
        super(SingleImage, self).__init__(image)

    def getImage(self):
        return self.images[0]


class AngledImageCollection(ImageCollection):
    def __init__(self, angleFn, *images):
        super(AngledImageCollection, self).__init__(*images)
        self.angleFn = angleFn

    def getImage(self):
        angle = (self.angleFn() + pi) % (2 * pi) - pi
        numImages = len(self.images)
        point = abs(int(angle * numImages / pi))
        if point == len(self.images):
            point -= 1
        image = self.images[point]
        return image


class Animation(ImageCollection):
    def __init__(self, speed, timeFunction, *images):
        super(Animation, self).__init__(*images)
        self.speed = speed
        self.timeFunction = timeFunction
        self.startNow()

    def getImage(self):
        elapsed = self.timeFunction() - self.start
        cycles = int(elapsed / self.speed)
        imgIndex = cycles % len(self.images)
        if elapsed > self.speed * len(self.images):
            self.start -= self.speed * len(self.images)
        return self.images[imgIndex]

    def startNow(self):
        self.start = self.timeFunction()

    def isComplete(self):
        elapsed = self.timeFunction() - self.start
        cycles = int(elapsed / self.speed)
        return cycles >= len(self.images)


class AnimationWithStart(Animation):
    '''Animation loop that starts with a set sequence (that doesn't loop),
    then goes into the main loop'''
    # NOTE: This class is untested. Remove this comment once you are certain
    #       it works correctly

    def __init__(self, speed, numInitial, timeFunction, *images):
        self.initial = True
        super(AnimationWithStart, self).__init__(
            speed, timeFunction, *images[numInitial - 1:])
        self.initialAnimation = Animation(speed, *images[:numInitial])

    def startNow(self):
        super(AnimationWithStart, self).startNow()
        self.initial = True

    def getImage(self):
        elapsed = self.timeFunction() - self.start
        cycles = int(elapsed / self.speed)
        imgIndex = cycles % len(self.images)
        if elapsed > self.speed * len(self.images):
            self.start -= self.speed * len(self.images)
            self.initial = False
        if self.initial:
            return self.initialAnimation.getImage()
        else:
            return self.images[imgIndex]



import logging
import os

from trosnoth.const import OPEN_CHAT
from trosnoth.data import getPath, user, makeDirs
from trosnoth.messages import (
    SetGameModeMsg, SetGameSpeedMsg, UpgradeApprovedMsg, AbandonUpgradeMsg,
    ChangeTimeLimitMsg, SetTeamNameMsg, UpgradeChangedMsg, TrosballTagMsg,
    ChatMsg, ZoneStateMsg, ReturnToLobbyMsg,
)
from trosnoth.model import modes
from trosnoth.model.player import Player
from trosnoth.model.upgrades import allUpgrades, upgradeOfType

log = logging.getLogger(__name__)


class GameModeHelper(object):
    def __init__(self, manholeHelper):
        self.manholeHelper = manholeHelper

    def __repr__(self):
        return 'GameModeHelper<%s>' % (self.manholeHelper.getGameMode(),)

    def __getattr__(self, name):
        if self.manholeHelper.getWorld().physics.hasMode(name):
            return lambda: self.manholeHelper.setGameMode(name)
        raise AttributeError(name)

    def __dir__(self):
        return self.manholeHelper.listGameModes()


class ManholeHelper(object):
    def __init__(self):
        self.mode = GameModeHelper(self)

    def getGame(self):
        raise NotImplementedError('getGame')

    def getWorld(self):
        return self.getGame().world

    def getBanner(self):
        game = self.getGame()
        if game is None:
            return ['\x1b[33mThe server is currently idle.']

        playerCount = sum(1 for p in self.getWorld().players if not p.bot)
        plural = '' if playerCount == 1 else 's'
        result = [
            '\x1b[32mA game is currently running.',
            '\x1b[102m\x1b[30m %d \x1b[49m\x1b[32m player%s online.' % (
                playerCount, plural),
        ]
        return result

    def getGameMode(self):
        '''
        Returns the current game mode.
        '''
        return self.getWorld().gameMode

    def setGameMode(self, gameMode):
        '''
        Changes the current game mode of the world.

        @returns: True if the specified game mode was valid, False otherwise.
        '''
        if self.getWorld().physics.hasMode(gameMode):
            self.getGame().sendServerCommand(SetGameModeMsg(gameMode.encode()))
            return True
        return False

    def listGameModes(self):
        '''
        Returns a list of valid game modes.
        '''
        return [m[len('setMode'):] for m in dir(modes.PhysicsConstants) if
                m.startswith('setMode') if m != 'setMode']

    def trosball(self):
        '''
        Sets the next game to be a Trosball game.
        '''
        self.setNextGameType('trosball')

    def setNextGameType(self, gameType):
        '''
        @param gameType: a string representing the type of game to start next
                time a game begins. For a normal game, use 'normal'.
        '''
        self.getWorld().setNextGameType(gameType)

    def getNextGameType(self):
        return self.getWorld().getNextGameType()

    def getSpeed(self):
        '''
        Returns the current speed of the world as a float, where 1.0 is
        the normal game speed.
        '''

        return self.getWorld()._gameSpeed

    def setSpeed(self, gameSpeed):
        '''
        Changes the speed of the world, where 1.0 is normal speed.

        @param gameSpeed: A positive number
        '''
        self.getGame().sendServerCommand(SetGameSpeedMsg(gameSpeed))

    def pause(self):
        self.setSpeed(0)

    def resume(self):
        self.setSpeed(1)

    def getPlayer(self, nick):
        if isinstance(nick, Player):
            return nick
        for p in self.getWorld().players:
            if p.user == nick or p.nick == nick:
                return p
        return None

    def kickPlayer(self, nick):
        p = self.getPlayer(nick)
        self.getGame().kickPlayer(p.id)

    def startGame(self):
        self.getWorld()._gameStarter.startNewGame()

    def giveUpgrade(self, player, upgradeCode):
        '''
        Gives an upgrade to a player, which is used immediately.

        @param player: Player, nick or username
        @param upgradeCode: Single character upgrade code
        '''
        player = self.getPlayer(player)
        player.pendingUpgrade = upgradeCode
        player.agent.messageToAgent(UpgradeApprovedMsg(upgradeCode))

    def removeUpgrade(self, player):
        '''
        Removes the specified player's current upgrade.

        @param player: Player, nick or username
        '''
        player = self.getPlayer(player)
        currentUpgrade = player.upgrade
        if currentUpgrade is not None:
            # TODO: this may leave the player out of sync, thinking that it
            # still has the upgrade
            self.getGame().sendServerCommand(
                AbandonUpgradeMsg(player.id, self.getWorld().lastTickId))

    def getTimeRemaining(self):
        '''
        Get the amount of time left on the game clock.
        '''
        return self.getWorld().getTimeLeft()

    def endNow(self):
        self.setTimeRemaining(1)

    def setTimeRemaining(self, time):
        '''
        Sets the remaining game time to the given number of seconds or, if None
        is given, to be unlimited.
        '''
        if time is None or time == 0:
            timeLimit = 0
        else:
            timeLimit = time

        self.getGame().sendServerCommand(ChangeTimeLimitMsg(timeLimit))

    def setPlayerLimits(self, maxPerTeam, maxTotal=0):
        '''
        Changes the player limits in the current game. Note that this does not
        affect players who are already in the game.

        @param maxPerTeam: Maximum number of players per team at once
        @param maxTotal: Maximum number of players in the game at once
        '''
        self.getGame().setPlayerLimits(maxPerTeam, maxTotal=maxTotal)

    def setTeamName(self, teamId, teamName):
        '''
        Changes the name of the specified team.

        @param teamId: The id of the team ('A' or 'B')
        @param teamName: The next team name
        '''
        try:
            team = self.getWorld().teamWithId[teamId]
        except KeyError:
            return
        self.getGame().sendServerCommand(
            SetTeamNameMsg(team.id, teamName.encode()))

    def getTeamNames(self):
        '''
        Returns the name of both teams as a tuple.
        '''
        return (
            self.getWorld().teams[0].teamName,
            self.getWorld().teams[1].teamName,
        )

    def getGameState(self):
        '''
        Returns the current state of the game as a string.

        @returns: The text id of the current game state
        '''
        return self.getGame().getGameState()

    def getWinningTeam(self):
        '''
        Returns the winning team.

        If the game hasn't finished yet, the behaviour is undefined.
        @returns: The ID of the winning team or None if it was a draw
        '''
        if self.getWorld().getWinningTeam() is None:
            return None

        return self.getWorld().getWinningTeam().id

    def getZoneCounts(self):
        '''
        Returns the number of zones owned by each team as a tuple.
        '''
        return (
            self.getWorld().teams[0].numZonesOwned,
            self.getWorld().teams[1].numZonesOwned,
        )

    def getUpgradeCosts(self):
        upgrades = {}

        for key, upgrade in upgradeOfType.iteritems():
            upgrades[key] = upgrade.requiredStars
        return upgrades

    def setUpgradeCosts(self, d):
        if isinstance(d, int):
            d = {k: d for k in upgradeOfType}
        for k, v in d.iteritems():
            self.setUpgradeCost(k, v)

    def getUpgrades(self):
        upgrades = []

        for key, upgrade in upgradeOfType.iteritems():
            newUpgrade = {
                "id": key,
                "name": upgrade.name,
                "starCost": upgrade.requiredStars,
                "timeLimit": upgrade.timeRemaining,
                "order": upgrade.order,
                "icon": "upgrade_blank",
                "special": False
            }

            if upgrade not in allUpgrades:
                newUpgrade["order"] += 1000
                newUpgrade["special"] = True

            if upgrade.iconPath is not None:
                newUpgrade["icon"] = upgrade.iconPath.replace("-", "_")[:-4]

            upgrades.append(newUpgrade)

        upgrades.sort(key=lambda upgrade: upgrade["order"])

        return upgrades

    def setUpgradeCost(self, upgradeId, cost):
        '''
        Changes how many stars are required to purchase a particular upgrade.

        @param upgradeId: The single character ID of the upgrade
        @param cost: The new number of stars required.
        '''
        if cost < 0:
            return

        for upgradeClass in allUpgrades:
            if upgradeClass.upgradeType == upgradeId:
                upgradeClass.requiredStars = cost
                break
        self.getGame().sendServerCommand(
            UpgradeChangedMsg(upgradeId, 'S', cost))

    def setUpgradeTime(self, upgradeId, time):
        '''
        Changes how long a particular upgrade lasts.

        @param upgradeId: The single character ID of the upgrade
        @param time: The time (in seconds) that the upgrade will last for.
            Setting this to 0 will make the upgrade never naturally run out.
        '''
        if time < 0:
            return

        for upgradeClass in allUpgrades:
            if upgradeClass.upgradeType == upgradeId:
                upgradeClass.timeRemaining = time
                break
        self.getGame().sendServerCommand(
            UpgradeChangedMsg(upgradeId, 'T', time))

    def fixTrosball(self, force=False):
        '''
        Resets the playing field during a game of Trosball.

        Used primarily to fix a stuck Trosball.
        @param force: If True, bypasses several sanity checks
        '''
        state = self.getWorld().state
        if not force:
            if not state.isTrosball():
                return "A Trosball game isn't running right now."
            if self.getWorld().trosballPlayer is not None:
                return '%s has the Trosball: is it really stuck?' % (
                    self.getWorld().trosballPlayer.nick,)

        self.getGame().sendServerCommand(TrosballTagMsg('\x00', '\x00'))

    def sendServerMessage(self, message):
        '''
        Sends a chat message from the server to all players.

        @param message: The message you want everyone to see.
        '''
        self.getGame().sendServerCommand(ChatMsg(OPEN_CHAT, text=message))

    def newWorld(
            self, halfMapWidth=None, mapHeight=None, gameDuration=None,
            blockRatio=0.5):
        if halfMapWidth is None:
            halfMapWidth = self.getWorld().map.layout.halfMapWidth
        if mapHeight is None:
            mapHeight = self.getWorld().map.layout.mapHeight
        layout = self.layoutDatabase.generateRandomMapLayout(
            halfMapWidth, mapHeight, blockRatio=blockRatio)
        self.getWorld().reset(layout, gameDuration)

    def loadMap(self, filename):
        '''
        Loads a map from the .trosnoth/maps directory.

        @param filename: The filename of the map (with extension)
        '''
        game = self.getGame()
        layout = game.layoutDatabase.loadMapFromFile(filename)
        game.world.loadedMap = layout

        return 'Map loaded; it will appear on the next game start.'

    def saveMap(self, filename, force=False):
        '''
        Saves the current map layout to the .trosnoth/maps directory.

        @param filename: The filename of the map (with extension)
        @param force: Overwrites the file if it already exists
        '''
        mapDir = getPath(user, 'maps')
        makeDirs(mapDir)
        filename = os.path.join(mapDir, filename)
        if os.path.exists(filename) and not force:
            return 'File already exists (use "force" parameter to overwrite)'

        with open(filename, 'w') as f:
            f.write(self.getWorld().layout.toString())
        return 'Map saved to %s' % (filename,)

    def ejectMap(self):
        '''
        Unloads any map that was previously loaded.
        '''
        self.getWorld().loadedMap = None

    def enableUpgrade(self, upgradeId):
        '''
        Turns the ability to buy the specified upgrade on.

        @param upgradeId: The single character ID of the upgrade
        '''
        for upgradeClass in allUpgrades:
            if upgradeClass.upgradeType == upgradeId:
                upgradeClass.enabled = True
        self.getGame().sendServerCommand(UpgradeChangedMsg(upgradeId, 'E', 1))

    def disableUpgrade(self, upgradeId):
        '''
        Turns the ability to buy the specified upgrade off.

        It can still be given via console and all existing upgrades will remain
        active until they expire normally.
        @param upgradeId: The single character ID of the upgrade
        '''
        for upgradeClass in allUpgrades:
            if upgradeClass.upgradeType == upgradeId:
                upgradeClass.enabled = False
        self.getGame().sendServerCommand(UpgradeChangedMsg(upgradeId, 'E', 0))

    def enableAllUpgrades(self):
        '''
        See documentation for enableUpgrade.
        '''
        for upgradeClass in allUpgrades:
            self.enableUpgrade(upgradeClass.upgradeType)

    def disableAllUpgrades(self):
        '''
        See documentation for disableUpgrade.
        '''
        for upgradeClass in allUpgrades:
            self.disableUpgrade(upgradeClass.upgradeType)

    def changeZoneOwners(self, state):
        '''
        Changes the owner of every zone on the map.

        @param state: A string representing the new owner of each zone.
        It should have a length equal to the number of zones, and the
        characters should be either a team ID, a dash (make neutral) or a
        question mark (don't alter this zone). Spaces are ignored, so you may
        include them for readability. All other characters are invalid.

        Zone counting starts at the top left, then goes top to bottom and
        left to right (in that order).
        '''
        if len(state) != len(self.getWorld().zoneWithId):
            return "Wrong number of zones (expected %d, got %d)" % \
                (len(self.getWorld().zoneWithId), len(state))

        state = state.replace(" ", "").replace("-", "\x00")

        validChars = "".join(self.getWorld().teamWithId.keys()) + "?"

        if state.strip(validChars):
            return "Invalid character in state string"

        for zone, team in enumerate(state):
            newTeam = team
            dark = True

            if team == "?":
                zoneObj = self.getWorld().zoneWithId[zone]
                newTeam = "\x00" if zoneObj.owner is None else zoneObj.owner.id
                dark = zoneObj.dark

            self.getGame().sendServerCommand(ZoneStateMsg(zone, newTeam, dark))

    def reset(self):
        self.getGame().sendServerCommand(ReturnToLobbyMsg())


class AuthServerManholeHelper(ManholeHelper):
    def __init__(self, authFactory):
        self.authFactory = authFactory
        super(AuthServerManholeHelper, self).__init__()

    def getGame(self):
        games = self.authFactory.servers.values()
        if len(games) == 0:
            log.info('getGame: No games available')
            return None
        return games[0].game


class LocalManholeHelper(ManholeHelper):
    def __init__(self, getGame):
        self.getGame = getGame
        super(LocalManholeHelper, self).__init__()



# Trosnoth (UberTweak Platform Game)
# Copyright (C) 2006-2012 Joshua D Bartlett
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# version 2 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301, USA.

from framework import Element, CompoundElement
from elements import TextButton, TextElement
from trosnoth.gui.common import (translateEvent, Location, addPositions,
    AttachedPoint, Relative)
from trosnoth.gui.errors import MultiWindowException
from trosnoth.gui.framework.utils import wrapWords
from trosnoth.utils.event import Event
from trosnoth.utils.utils import new
import pygame

class MoveableBox(CompoundElement):
    defaultBorderColour = (0,0,255)
    defaultTitleColour = (255,255,255)
    defaultBackgroundColour = (255,255,255)

    def __init__(self, app, size, caption, subCaption=None):
        CompoundElement.__init__(self, app)
        self.showing = False
        self.size = size
        self._edge = MoveableBoxEdge(self.app, self, caption, subCaption)

        self.setColours(self.defaultBorderColour, self.defaultTitleColour,
                self.defaultBackgroundColour)

        self.onClose = Event()

    def _giveFocus(self):
        self.app.screenManager.dialogFocus(self)

    def _getSize(self):
        if hasattr(self.size, 'getSize'):
            return self.size.getSize(self.app)
        else:
            return self.size

    def _setPos(self, pos):
        self._edge._setPos(pos)

    def setColours(self, borderColour=None, titleColour=None,
            backgroundColour=None):
        if borderColour:
            self._edge.borderColour = borderColour
        if titleColour:
            self._edge.titleColour = titleColour
        if backgroundColour:
            self.backgroundColour = backgroundColour

    def setCaption(self, caption=None, subCaption=None):
        if caption is not None:
            self._edge.caption = caption
        if subCaption is not None:
            if subCaption == False:
                self._edge.subCaption = None
            else:
                self._edge.subCaption = subCaption

    def _getPt(self):
        return self._edge._getInsideArea()

    def _getInsideRect(self):
        return pygame.Rect(self._edge._getInsideTopLeftPt(), self._getSize())

    def draw(self, screen):
        self._edge.draw(screen)
        subSurface = pygame.Surface(self._getSize())
        subSurface.fill(self.backgroundColour)
        CompoundElement.draw(self, subSurface)
        screen.blit(subSurface, self._edge._getInsideTopLeftPt())

    def processEvent(self, event):
        event = self._edge.processEvent(event)
        if event:
            if hasattr(event, 'pos'):
                event2 = translateEvent(event, self._edge._getInsideTopLeftPt())
                isPos = True
            else:
                event2 = event
                isPos = False
            event2 = CompoundElement.processEvent(self, event2)
            if event2 == None:
                return None
            elif isPos and self._edge._getEdgeRect().collidepoint(event.pos):
                return None
            else:
                return event

    def show(self):
        try:
            showDialog = self.app.screenManager.showDialog
        except AttributeError:
            raise MultiWindowException('Dialog Boxes cannot be used unless the '
                    'Application is a MultiWindowApplication')
        showDialog(self)
        self.showing = True

    def close(self):
        try:
            closeDialog = self.app.screenManager.closeDialog
        except AttributeError:
            raise MultiWindowException('Dialog Boxes cannot be used unless the '
                    'Application is a MultiWindowApplication')
        closeDialog(self)
        self.showing = False
        self.onClose.execute()

class DialogBox(MoveableBox):
    flashTime = 0.15
    flashNum = 2
    def __init__(self, app, size, caption, subCaption=None):
        MoveableBox.__init__(self, app, size, caption, subCaption)
        self.timeToFlash = 0

    def Relative(self, x, y):
        '''
        Returns an object representing a position relative to this dialogue box.
        See common.Relative() for more information.
        '''
        def getRect():
            return getInsideRectFromZero(self)
        return Relative(getRect, x, y)

    def processEvent(self, event):
        event = MoveableBox.processEvent(self, event)
        if event:
            if event.type == pygame.MOUSEBUTTONDOWN:
                # It was clicked, but not within us.
                self.timeToFlash = self.flashNum * self.flashTime * 2
                # (Times 2 because we flash on and off)

            # We want to hold any messages except quit ones
            if event.type == pygame.QUIT:
                return event
        return None

    def tick(self, deltaT):
        if self.timeToFlash > 0:
            self.timeToFlash = max(self.timeToFlash - deltaT, 0)
        MoveableBox.tick(self, deltaT)

    def draw(self, screen):
        if self.timeToFlash > 0:
            phase = int(self.timeToFlash / self.flashTime) % 2
            if (phase == 1 and self.__normalBorderColour ==
                    self._edge.borderColour):
                invert = lambda x: tuple([255-i for i in x])
                MoveableBox.setColours(self,
                        borderColour=invert(self.__normalBorderColour),
                        titleColour=invert(self.__normalTitleColour))
            elif (phase == 0 and self.__normalBorderColour !=
                    self._edge.borderColour):
                MoveableBox.setColours(self,
                        borderColour=self.__normalBorderColour,
                        titleColour=self.__normalTitleColour)
        MoveableBox.draw(self, screen)

    def setColours(self, borderColour=None, titleColour=None,
            backgroundColour=None):
        MoveableBox.setColours(self, borderColour, titleColour,
                backgroundColour)
        self.__normalBorderColour = self._edge.borderColour
        self.__normalTitleColour = self._edge.titleColour
        self.timeToFlash = 0

    def setCaption(self, caption=None, subCaption=None):
        MoveableBox.setCaption(self, caption, subCaption)

class MessageBox(DialogBox):
    '''buttons = a tuple of DialogResults'''
    def __init__(self, app, size, caption, message, buttons):
        DialogBox.__init__(self, app, size, caption)
        if hasattr(size, 'getSize'):
            size = size.getSize(app)
        textFont = app.screenManager.fonts.labelFont
        lines = wrapWords(app, message, textFont, size[0])
        y = 5
        for text in lines:
            textPos = Location(DialogBoxAttachedPoint(self, (0, y), 'midtop'),
                    'midtop')
            self.elements.append(TextElement(app, text, textFont, textPos))
            y += textFont.getHeight(app)
        buttonFont = app.screenManager.fonts.captionFont
        for x in range(0, len(buttons)):
            dialogResult = buttons[x]
            location = Location(DialogBoxAttachedPoint(self, (0,
                    -textFont.getHeight(self.app) / 2), 'midbottom'),
                    'midbottom')
        buttonFont = app.screenManager.fonts.captionFont
        for x in range(0, len(buttons)):
            dialogResult = buttons[x]
            location = Location(((x + 1) * size[0] / (len(buttons) + 1), size[1]
                    * 2 / 3), 'midtop')
            button = TextButton(app, location, DialogResult.text[dialogResult],
                    buttonFont, (255,0,0), (255,100,200))
            button._mbResult = dialogResult
            button.onClick.addListener(self._setResultAndClose)
            self.elements.append(button)
        self.result = None

    def _setResultAndClose(self, button):
        self.result = button._mbResult
        self.close()

class YesNoBox(MessageBox):
    def __init__(self, app, size, caption, message):
        buttons = (DialogResult.Yes, DialogResult.No)
        MessageBox.__init__(self, app, size, caption, message, buttons)

class YesNoCancelBox(MessageBox):
    def __init__(self, app, size, caption, message):
        buttons = (DialogResult.Yes, DialogResult.No, DialogResult.Cancel)
        MessageBox.__init__(self, app, size, caption, message, buttons)

class OkBox(MessageBox):
    def __init__(self, app, size, caption, message):
        buttons = tuple([DialogResult.OK])
        MessageBox.__init__(self, app, size, caption, message, buttons)

class OkCancelBox(MessageBox):
    def __init__(self, app, size, caption, message):
        buttons = (DialogResult.OK, DialogResult.Cancel)
        MessageBox.__init__(self, app, size, caption, message, buttons)

class OkRetryBox(MessageBox):
    def __init__(self, app, size, caption, message):
        buttons = (DialogResult.OK, DialogResult.Retry)
        MessageBox.__init__(self, app, size, caption, message, buttons)

class DialogResult(object):
    OK, Cancel, Yes, No, Retry = new(5)
    text = {OK:'OK', Cancel:'Cancel', Yes:'Yes', No:'No', Retry:'Retry'}


def getInsideRectFromZero(box):
    r = box._getInsideRect()
    r.topleft = (0,0)
    return r

def DialogBoxAttachedPoint(box, val, anchor='topleft'):
    return AttachedPoint(val, lambda:getInsideRectFromZero(box), anchor)

class MoveableBoxEdge(Element):
    '''
    Handles the border (including dragging and dropping) of boxes. It also
    handles the position of the box (so the box must query the edge to find its
    top left)
    '''
    def __init__(self, app, box, caption, subCaption=None):
        Element.__init__(self, app)
        self.box = box
        self.font = app.screenManager.fonts.captionFont
        self.smallFont = app.screenManager.fonts.keymapFont
        self.caption = caption
        self.subCaption = subCaption
        self.beingDragged = False
        self.borderWidth = 3
        appSize = app.screenManager.size
        self._setPos(tuple([(appSize[i] - self._getFullSize()[i]) / 2 for i
                in 0,1]))

    def _getFullSize(self):
        return (self._getFullWidth(), self.box._getSize()[1] +
                self.borderWidth + self._getHeaderHeight())

    def _setPos(self, pos):
        self.pos = pos

    def __movedBy(self, difference):
        self._setPos(addPositions(difference, self.pos))

    def _getInsideTopLeftPt(self):
        return addPositions(self.pos, (self.borderWidth,
                self._getHeaderHeight()))

    def _getHeaderHeight(self):
        return int(self.font.getLineSize(self.app) * 1.5)

    def _getFullWidth(self):
        return self.box._getSize()[0] + self.borderWidth * 2

    def _getHeaderRect(self):
        r = pygame.rect.Rect(self.pos, (self._getFullWidth(),
                self._getHeaderHeight()))
        return r

    def _getEdgeRect(self):
        return pygame.Rect(self.pos, (self._getFullSize()))

    def isPointInRect(self, pos):
        return self._getEdgeRect().collidepoint(pos)

    def isPointInHeader(self, pos):
        return self._getHeaderRect().collidepoint(pos)

    def processEvent(self, event):
        if (event.type == pygame.MOUSEBUTTONDOWN and
                self.isPointInRect(event.pos)):
            self.box._giveFocus()
        if (event.type == pygame.MOUSEBUTTONDOWN and event.button == 1 and
                self.isPointInHeader(event.pos)):
            self.beingDragged = True
            return None
        elif event.type == pygame.MOUSEMOTION and self.beingDragged:
            self.__movedBy(event.rel)
            return None
        elif (event.type == pygame.MOUSEBUTTONUP and event.button == 1 and
                self.beingDragged == True):
            self.beingDragged = False
            return None
        return event

    def draw(self, screen):
        screen.fill(self.borderColour, self._getEdgeRect())
        # Make pretty edges:
        setOuterBorder = lambda x, y: min(x + y, 255)
        outerBorder1 = tuple([setOuterBorder(self.borderColour[i], 70)
                for i in 0,1,2])
        outerBorder2 = tuple([setOuterBorder(self.borderColour[i], 35)
                for i in 0,1,2])
        pygame.draw.rect(screen, outerBorder1, self._getEdgeRect(), 1)
        if self.borderWidth > 2:
            oldRect = self._getEdgeRect()
            newRect = pygame.rect.Rect(oldRect.left + 1, oldRect.top + 1,
                    oldRect.width - 2, oldRect.height - 2)
            pygame.draw.rect(screen, outerBorder2, newRect, 1)

        text = self.font.render(self.app, self.caption, True,
                self.titleColour, self.borderColour)
        if self.subCaption is None:
            height = int(self.font.getLineSize(self.app) * 0.25)
        else:
            height = int(self.font.getLineSize(self.app) * 0.1)
        screen.blit(text, addPositions(self.pos, (self.borderWidth + 5,
                height)))
        if self.subCaption is not None:
            subText = self.smallFont.render(self.app, self.subCaption, True,
                    self.titleColour)
            height = int(self.font.getLineSize(self.app) * 0.8)
            screen.blit(subText, addPositions(self.pos, (self.borderWidth +
                    5, height)))




# Trosnoth (UberTweak Platform Game)
# Copyright (C) 2006-2011 Joshua D Bartlett
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# version 2 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301, USA.

import getpass
import os

from twisted.internet import defer

from trosnoth.run.main import Main
import trosnoth.main
from trosnoth.trosnothgui.pregame.authServerLoginBox import PasswordGUI
from trosnoth.trosnothgui.pregame.playscreen import PlayAuthScreen
from trosnoth.trosnothgui.pregame.serverSelectionScreen import (
        ServerSelectionScreen)
from trosnoth.trosnothgui.pregame.startup import StartupInterface
from trosnoth.trosnothgui.interface import SingleAuthInterface, Interface

class CampPasswordGetter(object):
    def getPassword(self, host, errorText=''):
        create = False
        username = getpass.getuser()
        password = username.encode('rot13')
        return defer.succeed((create, username, password))

class CampPlayAuthScreen(PlayAuthScreen):
    @staticmethod
    def passwordGUIFactory(app):
        if os.environ.get('LTSP_CLIENT'):
            return CampPasswordGetter()
        return PasswordGUI(app)


class CampSingleAuthInterface(SingleAuthInterface):
    connectScreenFactory = CampPlayAuthScreen

class CampServerSelectionScreen(ServerSelectionScreen):
    connectScreenFactory = CampPlayAuthScreen

class CampStartupInterface(StartupInterface):
    connectScreenFactory = CampPlayAuthScreen
    serverScreenFactory = CampServerSelectionScreen

class CampInterface(Interface):
    def _makeStartupInterface(self):
        return CampStartupInterface(self.app, self)

class CampMain(Main):
    singleInterfaceFactory = CampSingleAuthInterface
    standardInterfaceFactory = CampInterface

class main(trosnoth.main.main):
    appFactory = CampMain

if __name__ == '__main__':
    main()



# Trosnoth (UberTweak Platform Game)
# Copyright (C) 2006-2012 Joshua D Bartlett
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# version 2 as published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301, USA.


def distance(pt1, pt2):
    return ((pt1[0] - pt2[0]) ** 2 + (pt1[1] - pt2[1]) ** 2) ** 0.5


def fadeValues(val1, val2, interval):
    return val2 * interval + val1 * (1 - interval)


def fadeColours(c1, c2, interval):
    return (
        fadeValues(c1[0], c2[0], interval),
        fadeValues(c1[1], c2[1], interval),
        fadeValues(c1[2], c2[2], interval),
    )


def isNear(v1, v2, epsilon=1e-3):
    '''
    Compares two floating point values for approximate equality.
    '''
    return abs(v1 - v2) < epsilon
